from __future__ import annotations

from pathlib import Path
from types import SimpleNamespace

from click.testing import CliRunner
import pytest

from kanbus import cli
from test_helpers import build_project_configuration


def _run(args: list[str]) -> object:
    return CliRunner().invoke(cli.cli, args)


def test_hooks_list_and_validate_commands(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    monkeypatch.setattr(cli.Path, "cwd", lambda: tmp_path)

    monkeypatch.setattr(cli, "list_hooks", lambda _root: [])
    result_empty = _run(["hooks", "list"])
    assert result_empty.exit_code == 0
    assert "No hooks configured." in result_empty.output

    monkeypatch.setattr(
        cli,
        "list_hooks",
        lambda _root: [
            {
                "source": "external",
                "phase": "before",
                "event": "issue.create",
                "id": "h1",
                "command": "echo ok",
                "blocking": True,
                "timeout_ms": None,
            }
        ],
    )
    result_rows = _run(["hooks", "list"])
    assert result_rows.exit_code == 0
    assert "[external] before issue.create h1" in result_rows.output
    assert "timeout_ms: default" in result_rows.output

    monkeypatch.setattr(cli, "validate_hooks", lambda _root: ["bad hook"])  # type: ignore[arg-type]
    result_invalid = _run(["hooks", "validate"])
    assert result_invalid.exit_code != 0
    assert "Found 1 hook validation issue" in result_invalid.output

    monkeypatch.setattr(cli, "validate_hooks", lambda _root: [])
    result_valid = _run(["hooks", "validate"])
    assert result_valid.exit_code == 0
    assert "Hook configuration is valid." in result_valid.output


def test_wiki_render_and_list_commands(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    monkeypatch.setattr(cli.Path, "cwd", lambda: tmp_path)

    monkeypatch.setattr(cli, "render_wiki_page", lambda _req: "rendered")
    result_render = _run(["wiki", "render", "project/wiki/page.md"])
    assert result_render.exit_code == 0
    assert "rendered" in result_render.output

    monkeypatch.setattr(
        cli,
        "render_wiki_page",
        lambda _req: (_ for _ in ()).throw(cli.WikiError("wiki fail")),
    )
    result_render_fail = _run(["wiki", "render", "project/wiki/page.md"])
    assert result_render_fail.exit_code != 0
    assert "wiki fail" in result_render_fail.output

    monkeypatch.setattr(
        cli, "list_wiki_pages", lambda _root: ["project/wiki/a.md", "project/wiki/b.md"]
    )
    result_list = _run(["wiki", "list"])
    assert result_list.exit_code == 0
    assert "project/wiki/a.md" in result_list.output
    assert "project/wiki/b.md" in result_list.output

    monkeypatch.setattr(
        cli,
        "list_wiki_pages",
        lambda _root: (_ for _ in ()).throw(cli.WikiError("list fail")),
    )
    result_list_fail = _run(["wiki", "list"])
    assert result_list_fail.exit_code != 0
    assert "list fail" in result_list_fail.output


def test_edit_commands(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    monkeypatch.setattr(cli.Path, "cwd", lambda: tmp_path)

    monkeypatch.setattr(cli, "edit_view", lambda _root, _path, _rng: "view")
    assert _run(["edit", "view", "README.md", "--view-range", "1", "10"]).exit_code == 0

    monkeypatch.setattr(cli, "edit_str_replace", lambda _root, _path, _o, _n: "replace")
    assert (
        _run(
            ["edit", "str-replace", "README.md", "--old-str", "a", "--new-str", "b"]
        ).exit_code
        == 0
    )

    monkeypatch.setattr(cli, "edit_create", lambda _root, _path, _text: "create")
    assert _run(["edit", "create", "x.txt", "--file-text", "x"]).exit_code == 0

    monkeypatch.setattr(cli, "edit_insert", lambda _root, _path, _line, _text: "insert")
    assert (
        _run(
            ["edit", "insert", "x.txt", "--insert-line", "1", "--insert-text", "x"]
        ).exit_code
        == 0
    )

    monkeypatch.setattr(
        cli,
        "edit_view",
        lambda *_a: (_ for _ in ()).throw(cli.TextEditorError("view fail")),
    )
    result_fail = _run(["edit", "view", "README.md"])
    assert result_fail.exit_code != 0
    assert "view fail" in result_fail.output

    monkeypatch.setattr(
        cli,
        "edit_str_replace",
        lambda *_a: (_ for _ in ()).throw(cli.TextEditorError("replace fail")),
    )
    result_replace_fail = _run(
        ["edit", "str-replace", "README.md", "--old-str", "a", "--new-str", "b"]
    )
    assert result_replace_fail.exit_code != 0
    assert "replace fail" in result_replace_fail.output

    monkeypatch.setattr(
        cli,
        "edit_create",
        lambda *_a: (_ for _ in ()).throw(cli.TextEditorError("create fail")),
    )
    result_create_fail = _run(["edit", "create", "x.txt", "--file-text", "x"])
    assert result_create_fail.exit_code != 0
    assert "create fail" in result_create_fail.output

    monkeypatch.setattr(
        cli,
        "edit_insert",
        lambda *_a: (_ for _ in ()).throw(cli.TextEditorError("insert fail")),
    )
    result_insert_fail = _run(
        ["edit", "insert", "x.txt", "--insert-line", "1", "--insert-text", "x"]
    )
    assert result_insert_fail.exit_code != 0
    assert "insert fail" in result_insert_fail.output


def test_policy_check_command_paths(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    monkeypatch.setattr(cli.Path, "cwd", lambda: tmp_path)

    lookup = SimpleNamespace(
        issue=SimpleNamespace(identifier="kanbus-1"), project_dir=tmp_path / "project"
    )

    monkeypatch.setattr(
        "kanbus.issue_lookup.load_issue_from_project", lambda _root, _id: lookup
    )
    monkeypatch.setattr(
        "kanbus.config_loader.load_project_configuration", lambda _p: SimpleNamespace()
    )
    monkeypatch.setattr(
        "kanbus.project.get_configuration_path", lambda _p: tmp_path / ".kanbus.yml"
    )

    policies = lookup.project_dir / "policies"
    policies.mkdir(parents=True)

    monkeypatch.setattr("kanbus.policy_loader.load_policies", lambda _p: [])
    result_no_files = _run(["policy", "check", "kanbus-1"])
    assert result_no_files.exit_code == 0
    assert "No policy files found" in result_no_files.output

    monkeypatch.setattr(
        "kanbus.policy_loader.load_policies", lambda _p: [("p.feature", object())]
    )
    monkeypatch.setattr(
        "kanbus.issue_listing.load_issues_from_directory", lambda _d: []
    )
    monkeypatch.setattr(
        "kanbus.policy_evaluator.evaluate_policies_with_options", lambda *_a, **_k: []
    )

    result_ok = _run(["policy", "check", "kanbus-1"])
    assert result_ok.exit_code == 0
    assert "All policies passed for kanbus-1" in result_ok.output

    monkeypatch.setattr(
        "kanbus.issue_lookup.load_issue_from_project", lambda _root, _id: lookup
    )
    (lookup.project_dir / "policies").rmdir()
    result_no_dir = _run(["policy", "check", "kanbus-1"])
    assert result_no_dir.exit_code == 0
    assert "No policies directory found" in result_no_dir.output
    policies.mkdir(parents=True)

    monkeypatch.setattr(
        "kanbus.policy_evaluator.evaluate_policies_with_options",
        lambda *_a, **_k: [RuntimeError("v1")],
    )
    result_violation = _run(["policy", "check", "kanbus-1"])
    assert result_violation.exit_code != 0
    assert (
        "policy violation" in result_violation.output.lower()
        or "Found 1 policy violation" in result_violation.output
    )

    monkeypatch.setattr(
        "kanbus.issue_lookup.load_issue_from_project",
        lambda *_a, **_k: (_ for _ in ()).throw(RuntimeError("lookup fail")),
    )
    result_lookup_fail = _run(["policy", "check", "kanbus-1"])
    assert result_lookup_fail.exit_code != 0
    assert "lookup fail" in result_lookup_fail.output

    monkeypatch.setattr(
        "kanbus.issue_lookup.load_issue_from_project",
        lambda *_a, **_k: (_ for _ in ()).throw(cli.IssueLookupError("not found")),
    )
    result_lookup_issue_fail = _run(["policy", "check", "kanbus-1"])
    assert result_lookup_issue_fail.exit_code != 0
    assert "not found" in result_lookup_issue_fail.output


def test_policy_guide_list_steps_validate_commands(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    monkeypatch.setattr(cli.Path, "cwd", lambda: tmp_path)

    lookup = SimpleNamespace(
        issue=SimpleNamespace(identifier="kanbus-1"), project_dir=tmp_path / "project"
    )
    monkeypatch.setattr(
        "kanbus.issue_lookup.load_issue_from_project", lambda _root, _id: lookup
    )

    report = SimpleNamespace(violations=[], guidance_items=[])
    monkeypatch.setattr(
        "kanbus.policy_guidance.collect_guidance_for_issue", lambda **_k: report
    )
    monkeypatch.setattr(
        "kanbus.policy_guidance.sorted_deduped_guidance_items", lambda items: items
    )

    result_none = _run(["policy", "guide", "kanbus-1"])
    assert result_none.exit_code == 0
    assert "No guidance for kanbus-1" in result_none.output

    report.guidance_items = [
        SimpleNamespace(
            severity=SimpleNamespace(value="warning"),
            message="warn",
            explanations=["exp"],
        )
    ]
    result_guidance = _run(["policy", "guide", "kanbus-1"])
    assert result_guidance.exit_code == 0
    assert "GUIDANCE WARNING: warn" in result_guidance.output

    report.violations = [RuntimeError("bad policy")]
    result_validation = _run(["policy", "guide", "kanbus-1"])
    assert result_validation.exit_code != 0
    assert "validation issue" in result_validation.output.lower()

    monkeypatch.setattr(
        "kanbus.issue_lookup.load_issue_from_project",
        lambda *_a, **_k: (_ for _ in ()).throw(RuntimeError("guide fail")),
    )
    result_guide_fail = _run(["policy", "guide", "kanbus-1"])
    assert result_guide_fail.exit_code != 0
    assert "guide fail" in result_guide_fail.output

    monkeypatch.setattr(
        "kanbus.issue_lookup.load_issue_from_project",
        lambda *_a, **_k: (_ for _ in ()).throw(cli.IssueLookupError("no issue")),
    )
    result_guide_issue_fail = _run(["policy", "guide", "kanbus-1"])
    assert result_guide_issue_fail.exit_code != 0
    assert "no issue" in result_guide_issue_fail.output
    monkeypatch.setattr(
        "kanbus.issue_lookup.load_issue_from_project", lambda _root, _id: lookup
    )

    monkeypatch.setattr(
        "kanbus.project.load_project_directory", lambda _r: tmp_path / "project"
    )
    policies = tmp_path / "project" / "policies"
    policies.mkdir(parents=True, exist_ok=True)

    doc = SimpleNamespace(
        feature=SimpleNamespace(
            name="F", children=[SimpleNamespace(scenario=SimpleNamespace(name="S"))]
        )
    )
    monkeypatch.setattr(
        "kanbus.policy_loader.load_policies", lambda _p: [("p.feature", doc)]
    )

    monkeypatch.setattr(
        "kanbus.project.load_project_directory",
        lambda _r: tmp_path / "project-no-policies-dir",
    )
    (tmp_path / "project-no-policies-dir").mkdir(parents=True, exist_ok=True)
    result_list_no_dir = _run(["policy", "list"])
    assert result_list_no_dir.exit_code == 0
    assert "No policies directory found" in result_list_no_dir.output

    monkeypatch.setattr(
        "kanbus.project.load_project_directory", lambda _r: tmp_path / "project"
    )
    result_list = _run(["policy", "list"])
    assert result_list.exit_code == 0
    assert "p.feature" in result_list.output
    assert "Feature: F" in result_list.output
    assert "Scenario: S" in result_list.output

    doc_with_rule = SimpleNamespace(
        feature=SimpleNamespace(
            name="F2",
            children=[
                SimpleNamespace(
                    rule=SimpleNamespace(
                        name="R1",
                        children=[
                            SimpleNamespace(scenario=SimpleNamespace(name="Nested"))
                        ],
                    )
                )
            ],
        )
    )
    monkeypatch.setattr(
        "kanbus.policy_loader.load_policies",
        lambda _p: [("rule.feature", doc_with_rule)],
    )
    result_list_rule = _run(["policy", "list"])
    assert result_list_rule.exit_code == 0
    assert "Rule: R1 / Nested" in result_list_rule.output

    doc_skip_paths = SimpleNamespace(
        feature=SimpleNamespace(
            name="F3",
            children=[
                SimpleNamespace(other="ignored"),
                SimpleNamespace(
                    rule=SimpleNamespace(
                        name="R2",
                        children=[SimpleNamespace(other="ignored")],
                    )
                ),
            ],
        )
    )
    monkeypatch.setattr(
        "kanbus.policy_loader.load_policies",
        lambda _p: [("skip.feature", doc_skip_paths)],
    )
    result_skip_paths = _run(["policy", "list"])
    assert result_skip_paths.exit_code == 0
    assert "skip.feature" in result_skip_paths.output

    monkeypatch.setattr("kanbus.policy_loader.load_policies", lambda _p: [])
    result_list_none = _run(["policy", "list"])
    assert result_list_none.exit_code == 0
    assert "No policy files found" in result_list_none.output

    monkeypatch.setattr(
        "kanbus.project.load_project_directory",
        lambda *_a, **_k: (_ for _ in ()).throw(RuntimeError("list fail")),
    )
    result_list_fail = _run(["policy", "list"])
    assert result_list_fail.exit_code != 0
    assert "list fail" in result_list_fail.output

    steps = [
        SimpleNamespace(
            category=SimpleNamespace(value="Given"),
            description="D1",
            usage_pattern="P1",
        ),
        SimpleNamespace(
            category=SimpleNamespace(value="Then"), description="D2", usage_pattern="P2"
        ),
    ]
    monkeypatch.setattr(
        "kanbus.policy_evaluator._get_step_registry",
        lambda: SimpleNamespace(steps=steps),
    )

    result_steps = _run(["policy", "steps", "--category", "given", "--search", "d1"])
    assert result_steps.exit_code == 0
    assert "Given - D1" in result_steps.output

    result_steps_none = _run(["policy", "steps", "--category", "when"])
    assert result_steps_none.exit_code == 0
    assert "No matching steps found" in result_steps_none.output

    result_steps_search_none = _run(["policy", "steps", "--search", "zzz"])
    assert result_steps_search_none.exit_code == 0
    assert "No matching steps found" in result_steps_search_none.output

    monkeypatch.setattr(
        "kanbus.project.load_project_directory", lambda _r: tmp_path / "project"
    )
    monkeypatch.setattr(
        "kanbus.policy_loader.load_policies", lambda _p: [("p.feature", doc)]
    )
    monkeypatch.setattr(
        cli, "load_project_configuration", lambda _p: build_project_configuration()
    )
    monkeypatch.setattr(
        cli, "get_configuration_path", lambda _p: tmp_path / ".kanbus.yml"
    )
    monkeypatch.setattr(
        "kanbus.policy_evaluator.validate_policy_documents", lambda *_a: []
    )
    result_validate_ok = _run(["policy", "validate"])
    assert result_validate_ok.exit_code == 0
    assert "All 1 policy files are valid" in result_validate_ok.output

    monkeypatch.setattr(
        "kanbus.project.load_project_directory",
        lambda _r: tmp_path / "project-no-policies",
    )
    (tmp_path / "project-no-policies").mkdir(parents=True, exist_ok=True)
    result_validate_no_dir = _run(["policy", "validate"])
    assert result_validate_no_dir.exit_code == 0
    assert "No policies directory found" in result_validate_no_dir.output

    monkeypatch.setattr(
        "kanbus.project.load_project_directory", lambda _r: tmp_path / "project"
    )
    monkeypatch.setattr("kanbus.policy_loader.load_policies", lambda _p: [])
    result_validate_none = _run(["policy", "validate"])
    assert result_validate_none.exit_code == 0
    assert "No policy files found" in result_validate_none.output

    monkeypatch.setattr(
        "kanbus.policy_loader.load_policies", lambda _p: [("p.feature", doc)]
    )
    monkeypatch.setattr(
        "kanbus.policy_evaluator.validate_policy_documents",
        lambda *_a: [RuntimeError("v")],
    )
    result_validate_fail = _run(["policy", "validate"])
    assert result_validate_fail.exit_code != 0
    assert "validation issue" in result_validate_fail.output.lower()



from __future__ import annotations

from pathlib import Path
from types import SimpleNamespace

import pytest

from kanbus import daemon_client
from kanbus.daemon_protocol import ErrorEnvelope, ResponseEnvelope


def ok_response(request_id: str, result: dict | None = None) -> ResponseEnvelope:
    return ResponseEnvelope(
        protocol_version="1.0",
        request_id=request_id,
        status="ok",
        result=result or {},
    )


def error_response(request_id: str, message: str) -> ResponseEnvelope:
    return ResponseEnvelope(
        protocol_version="1.0",
        request_id=request_id,
        status="error",
        error=ErrorEnvelope(code="internal_error", message=message, details={}),
    )


def test_is_daemon_enabled_env_matrix(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.delenv("KANBUS_NO_DAEMON", raising=False)
    assert daemon_client.is_daemon_enabled() is True
    for value in ["1", "true", "yes", "TRUE", "Yes"]:
        monkeypatch.setenv("KANBUS_NO_DAEMON", value)
        assert daemon_client.is_daemon_enabled() is False
    monkeypatch.setenv("KANBUS_NO_DAEMON", "0")
    assert daemon_client.is_daemon_enabled() is True


def test_request_index_list_happy_path_and_disabled(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    root = tmp_path
    socket_path = root / "sock"
    socket_path.write_text("", encoding="utf-8")
    monkeypatch.setattr(daemon_client, "get_daemon_socket_path", lambda _r: socket_path)
    monkeypatch.setattr(
        daemon_client,
        "spawn_daemon",
        lambda _r: (_ for _ in ()).throw(RuntimeError("should not spawn")),
    )
    monkeypatch.setattr(
        daemon_client,
        "_request_with_recovery",
        lambda _s, request, _r: ok_response(
            request.request_id, {"issues": [{"id": "kanbus-1"}]}
        ),
    )
    monkeypatch.delenv("KANBUS_NO_DAEMON", raising=False)

    issues = daemon_client.request_index_list(root)
    assert issues == [{"id": "kanbus-1"}]

    monkeypatch.setenv("KANBUS_NO_DAEMON", "1")
    with pytest.raises(daemon_client.DaemonClientError, match="daemon disabled"):
        daemon_client.request_index_list(root)


def test_request_index_list_spawns_when_socket_missing(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    root = tmp_path
    socket_path = root / "missing.sock"
    spawned: dict[str, bool] = {}
    monkeypatch.setattr(daemon_client, "get_daemon_socket_path", lambda _r: socket_path)
    monkeypatch.setattr(
        daemon_client,
        "spawn_daemon",
        lambda _r: spawned.setdefault("spawned", True),
    )
    monkeypatch.setattr(
        daemon_client,
        "_request_with_recovery",
        lambda _s, request, _r: ok_response(request.request_id, {"issues": []}),
    )
    monkeypatch.delenv("KANBUS_NO_DAEMON", raising=False)
    daemon_client.request_index_list(root)
    assert spawned.get("spawned") is True


def test_request_index_list_error_response_uses_envelope_message(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    root = tmp_path
    socket_path = root / "sock"
    socket_path.write_text("", encoding="utf-8")
    monkeypatch.setattr(daemon_client, "get_daemon_socket_path", lambda _r: socket_path)
    monkeypatch.setattr(
        daemon_client,
        "_request_with_recovery",
        lambda _s, request, _r: error_response(request.request_id, "daemon boom"),
    )
    monkeypatch.delenv("KANBUS_NO_DAEMON", raising=False)
    with pytest.raises(daemon_client.DaemonClientError, match="daemon boom"):
        daemon_client.request_index_list(root)


def test_request_status_and_shutdown_paths(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    root = tmp_path
    socket_path = root / "sock"
    socket_path.write_text("", encoding="utf-8")
    monkeypatch.setattr(daemon_client, "get_daemon_socket_path", lambda _r: socket_path)

    def responder(_s: Path, request: object, _r: Path) -> ResponseEnvelope:
        req = request  # type: ignore[assignment]
        if req.action == "ping":
            return ok_response(req.request_id, {"status": "ok"})
        return ok_response(req.request_id, {"stopped": True})

    monkeypatch.setattr(daemon_client, "_request_with_recovery", responder)
    monkeypatch.delenv("KANBUS_NO_DAEMON", raising=False)
    assert daemon_client.request_status(root) == {"status": "ok"}
    assert daemon_client.request_shutdown(root) == {"stopped": True}

    monkeypatch.setenv("KANBUS_NO_DAEMON", "true")
    with pytest.raises(daemon_client.DaemonClientError, match="daemon disabled"):
        daemon_client.request_status(root)
    with pytest.raises(daemon_client.DaemonClientError, match="daemon disabled"):
        daemon_client.request_shutdown(root)


def test_request_status_and_shutdown_error_fallback_message(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    root = tmp_path
    socket_path = root / "sock"
    socket_path.write_text("", encoding="utf-8")
    monkeypatch.setattr(daemon_client, "get_daemon_socket_path", lambda _r: socket_path)
    monkeypatch.setattr(
        daemon_client,
        "_request_with_recovery",
        lambda _s, request, _r: ResponseEnvelope(
            protocol_version="1.0",
            request_id=request.request_id,
            status="error",
            error=None,
        ),
    )
    monkeypatch.delenv("KANBUS_NO_DAEMON", raising=False)
    with pytest.raises(daemon_client.DaemonClientError, match="daemon error"):
        daemon_client.request_status(root)
    with pytest.raises(daemon_client.DaemonClientError, match="daemon error"):
        daemon_client.request_shutdown(root)


def test_request_with_recovery_non_connection_error_passthrough(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    socket_path = tmp_path / "sock"
    request = SimpleNamespace(request_id="req-1")
    monkeypatch.setattr(
        daemon_client,
        "send_request",
        lambda *_a: (_ for _ in ()).throw(
            daemon_client.DaemonClientError("empty daemon response")
        ),
    )
    with pytest.raises(daemon_client.DaemonClientError, match="empty daemon response"):
        daemon_client._request_with_recovery(socket_path, request, tmp_path)


def test_request_with_recovery_retries_after_connection_failure(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    socket_path = tmp_path / "sock"
    socket_path.write_text("", encoding="utf-8")
    request = SimpleNamespace(request_id="req-2")
    attempts = {"count": 0}
    spawned: dict[str, bool] = {}

    def flaky_send(_s: Path, _r: object) -> ResponseEnvelope:
        attempts["count"] += 1
        if attempts["count"] < 3:
            raise daemon_client.DaemonClientError("daemon connection failed")
        return ok_response("req-2", {"ok": True})

    monkeypatch.setattr(daemon_client, "send_request", flaky_send)
    monkeypatch.setattr(
        daemon_client,
        "spawn_daemon",
        lambda _r: spawned.setdefault("spawned", True),
    )
    monkeypatch.setattr(daemon_client.time, "sleep", lambda *_a: None)

    response = daemon_client._request_with_recovery(socket_path, request, tmp_path)
    assert response.status == "ok"
    assert spawned.get("spawned") is True
    assert attempts["count"] >= 3
    assert not socket_path.exists()


def test_request_with_recovery_exhausts_retries(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    socket_path = tmp_path / "sock"
    request = SimpleNamespace(request_id="req-3")
    monkeypatch.setattr(
        daemon_client,
        "send_request",
        lambda *_a: (_ for _ in ()).throw(
            daemon_client.DaemonClientError("daemon connection failed")
        ),
    )
    monkeypatch.setattr(daemon_client, "spawn_daemon", lambda _r: None)
    monkeypatch.setattr(daemon_client.time, "sleep", lambda *_a: None)
    with pytest.raises(
        daemon_client.DaemonClientError, match="daemon connection failed"
    ):
        daemon_client._request_with_recovery(socket_path, request, tmp_path)


def test_send_request_success_empty_and_connection_failures(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    socket_path = tmp_path / "sock"
    request = daemon_client.RequestEnvelope(
        protocol_version="1.0",
        request_id="req-1",
        action="ping",
        payload={},
    )

    class FakeFile:
        def __init__(self, data: bytes):
            self._data = data

        def readline(self) -> bytes:
            return self._data

    class FakeSocket:
        def __init__(self, response: bytes):
            self.sent = b""
            self.response = response
            self.connected = None

        def settimeout(self, _timeout: float) -> None:
            pass

        def connect(self, path: str) -> None:
            self.connected = path

        def sendall(self, payload: bytes) -> None:
            self.sent = payload

        def makefile(self, _mode: str):
            return FakeFile(self.response)

        def __enter__(self):
            return self

        def __exit__(self, *_args):
            return False

    fake = FakeSocket(
        b'{"protocol_version":"1.0","request_id":"req-1","status":"ok","result":{"ok":true}}\n'
    )
    monkeypatch.setattr(daemon_client.socket, "socket", lambda *_a: fake)
    response = daemon_client.send_request(socket_path, request)
    assert response.status == "ok"
    assert response.result == {"ok": True}
    assert fake.connected == str(socket_path)
    assert fake.sent.endswith(b"\n")

    empty = FakeSocket(b"")
    monkeypatch.setattr(daemon_client.socket, "socket", lambda *_a: empty)
    with pytest.raises(daemon_client.DaemonClientError, match="empty daemon response"):
        daemon_client.send_request(socket_path, request)

    def raising_socket(*_a):
        raise OSError("socket down")

    monkeypatch.setattr(daemon_client.socket, "socket", raising_socket)
    with pytest.raises(
        daemon_client.DaemonClientError, match="daemon connection failed"
    ):
        daemon_client.send_request(socket_path, request)


def test_spawn_daemon_invokes_subprocess(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    root = tmp_path
    captured: dict[str, object] = {}

    def fake_popen(cmd, **kwargs):
        captured["cmd"] = cmd
        captured["kwargs"] = kwargs
        return None

    monkeypatch.setattr(daemon_client.subprocess, "Popen", fake_popen)
    daemon_client.spawn_daemon(root)
    assert captured["cmd"][0] == daemon_client.sys.executable
    assert captured["cmd"][1:3] == ["-m", "kanbus.daemon"]
    assert captured["kwargs"]["cwd"] == root
    assert captured["kwargs"]["start_new_session"] is True


def test_request_with_recovery_raises_non_connection_error_during_retry(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    socket_path = tmp_path / "sock"
    request = SimpleNamespace(request_id="req-4")
    state = {"count": 0}

    def flaky_send(_s: Path, _r: object):
        state["count"] += 1
        if state["count"] == 1:
            raise daemon_client.DaemonClientError("daemon connection failed")
        raise daemon_client.DaemonClientError("empty daemon response")

    monkeypatch.setattr(daemon_client, "send_request", flaky_send)
    monkeypatch.setattr(daemon_client, "spawn_daemon", lambda _r: None)
    monkeypatch.setattr(daemon_client.time, "sleep", lambda *_a: None)
    with pytest.raises(daemon_client.DaemonClientError, match="empty daemon response"):
        daemon_client._request_with_recovery(socket_path, request, tmp_path)



"""Policy guidance hook execution and rendering."""

from __future__ import annotations

import os
from pathlib import Path

import click

from kanbus.config_loader import load_project_configuration
from kanbus.issue_listing import load_issues_from_directory
from kanbus.models import IssueData
from kanbus.policy_context import (
    GuidanceItem,
    GuidanceSeverity,
    PolicyContext,
    PolicyEvaluationReport,
    PolicyOperation,
)
from kanbus.policy_evaluator import (
    PolicyEvaluationMode,
    PolicyEvaluationOptions,
    evaluate_policies_report,
)
from kanbus.policy_loader import load_policies
from kanbus.project import get_configuration_path, load_project_directory


def guidance_is_enabled(no_guidance: bool = False) -> bool:
    """Return True when guidance emission is enabled."""
    if no_guidance:
        return False
    raw = os.environ.get("KANBUS_NO_GUIDANCE", "").strip().lower()
    return raw not in {"1", "true", "yes", "on"}


def collect_guidance_for_issue(
    root: Path,
    issue: IssueData,
    operation: PolicyOperation,
) -> PolicyEvaluationReport:
    """Evaluate policies in guidance mode for one issue."""
    project_dir = load_project_directory(root)
    policies_dir = project_dir / "policies"
    if not policies_dir.is_dir():
        return PolicyEvaluationReport()

    policy_documents = load_policies(policies_dir)
    if not policy_documents:
        return PolicyEvaluationReport()

    configuration = load_project_configuration(get_configuration_path(project_dir))
    issues_dir = project_dir / "issues"
    all_issues = load_issues_from_directory(issues_dir)

    context = PolicyContext(
        current_issue=issue,
        proposed_issue=issue,
        transition=None,
        operation=operation,
        project_configuration=configuration,
        all_issues=all_issues,
    )

    return evaluate_policies_report(
        context,
        policy_documents,
        PolicyEvaluationOptions(
            collect_all_violations=True,
            mode=PolicyEvaluationMode.GUIDANCE,
        ),
    )


def emit_guidance_for_issues(
    root: Path,
    issues: list[IssueData],
    operation: PolicyOperation,
    *,
    no_guidance: bool = False,
) -> None:
    """Run non-blocking guidance hooks and emit results on stderr."""
    if not guidance_is_enabled(no_guidance):
        return

    collected: list[GuidanceItem] = []
    for issue in issues:
        try:
            report = collect_guidance_for_issue(root, issue, operation)
        except Exception:
            # Guidance must never block successful command paths.
            continue

        collected.extend(report.guidance_items)
        for violation in report.violations:
            collected.append(
                GuidanceItem(
                    severity=GuidanceSeverity.WARNING,
                    message=(
                        "Guidance policy error "
                        f"({violation.policy_file} / Rule: {violation.scenario}): {violation.message}"
                    ),
                    explanations=[
                        'Explanation: Run "kbs policy validate" to fix this policy definition.'
                    ],
                    policy_file=violation.policy_file,
                    scenario=violation.scenario,
                    step=violation.failed_step,
                )
            )

    for item in sorted_deduped_guidance_items(collected):
        prefix = (
            "GUIDANCE WARNING"
            if item.severity == GuidanceSeverity.WARNING
            else "GUIDANCE SUGGESTION"
        )
        click.echo(f"{prefix}: {item.message}", err=True)
        for explanation in item.explanations:
            if explanation.startswith("Explanation:"):
                click.echo(f"  {explanation}", err=True)
            else:
                click.echo(f"  Explanation: {explanation}", err=True)


def sorted_deduped_guidance_items(items: list[GuidanceItem]) -> list[GuidanceItem]:
    seen: set[tuple[str, str, tuple[str, ...], str, str, str]] = set()
    unique: list[GuidanceItem] = []
    for item in items:
        key = (
            item.severity.value,
            item.message,
            tuple(item.explanations),
            item.policy_file,
            item.scenario,
            item.step,
        )
        if key in seen:
            continue
        seen.add(key)
        unique.append(item)

    return sorted(
        unique,
        key=lambda item: (
            0 if item.severity == GuidanceSeverity.WARNING else 1,
            item.policy_file,
            item.scenario,
            item.step,
            item.message,
        ),
    )



from __future__ import annotations

from pathlib import Path
from unittest.mock import patch

from kanbus import ai_summarize
from kanbus.models import AiConfiguration


def _ai_config() -> AiConfiguration:
    return AiConfiguration(provider="test", model="gpt-test")


def test_cache_key_uses_identifier_updated_at_and_detail() -> None:
    issue = {"id": "kanbus-1", "updated_at": "2026-01-01T00:00:00Z"}
    key_a = ai_summarize._cache_key(issue, "short")
    key_b = ai_summarize._cache_key(issue, "long")
    assert key_a != key_b


def test_read_cache_returns_none_for_missing_and_invalid_json(tmp_path: Path) -> None:
    assert ai_summarize._read_cache(tmp_path, "k") is None
    (tmp_path / ai_summarize.AI_SUMMARIES_CACHE).write_text("{bad", encoding="utf-8")
    assert ai_summarize._read_cache(tmp_path, "k") is None


def test_write_cache_persists_and_updates_entries(tmp_path: Path) -> None:
    ai_summarize._write_cache(tmp_path, "k1", "v1")
    ai_summarize._write_cache(tmp_path, "k2", "v2")
    assert ai_summarize._read_cache(tmp_path, "k1") == "v1"
    assert ai_summarize._read_cache(tmp_path, "k2") == "v2"


def test_summarize_issue_returns_not_configured_when_ai_absent() -> None:
    issue = {"identifier": "kanbus-2", "title": "Title"}
    assert (
        ai_summarize.summarize_issue(issue, "short", None)
        == "(AI summarization not configured)"
    )


def test_make_ai_summarize_handles_missing_issue_and_cache_hits(
    tmp_path: Path, monkeypatch
) -> None:
    monkeypatch.delenv("KANBUS_TEST_AI_MOCK", raising=False)
    issue = {"identifier": "kanbus-3", "title": "A title", "updated_at": "2026-01-01"}
    fn = ai_summarize.make_ai_summarize(
        {"kanbus-3": issue},
        _ai_config(),
        tmp_path,
    )

    assert fn("missing") == "(issue not found)"
    first = fn("kanbus-3", "short")
    second = fn("kanbus-3", "short")
    assert first == "Summary: A title"
    assert second == first


def test_make_ai_summarize_mock_mode_logs_calls(tmp_path: Path, monkeypatch) -> None:
    monkeypatch.setenv("KANBUS_TEST_AI_MOCK", "1")
    issue = {"identifier": "kanbus-4", "title": "B title", "updated_at": "2026-01-02"}
    fn = ai_summarize.make_ai_summarize(
        {"kanbus-4": issue},
        _ai_config(),
        tmp_path,
    )
    result = fn("kanbus-4", "short")
    assert result == "Generated summary for kanbus-4"
    log_path = tmp_path / ai_summarize.AI_CALLS_LOG
    assert log_path.exists()
    assert log_path.read_text(encoding="utf-8").strip() == "1"


def test_cache_key_falls_back_to_identifier_and_empty_updated_at() -> None:
    issue = {"identifier": "kanbus-5"}
    key = ai_summarize._cache_key(issue, "short")
    assert isinstance(key, str)
    assert len(key) == 64


def test_read_cache_handles_oserror_when_reading(tmp_path: Path) -> None:
    path = tmp_path / ai_summarize.AI_SUMMARIES_CACHE
    path.write_text("{}", encoding="utf-8")
    with patch.object(Path, "read_text", side_effect=OSError("boom")):
        assert ai_summarize._read_cache(tmp_path, "k") is None


def test_write_cache_tolerates_invalid_existing_json(tmp_path: Path) -> None:
    (tmp_path / ai_summarize.AI_SUMMARIES_CACHE).write_text("{bad", encoding="utf-8")
    ai_summarize._write_cache(tmp_path, "k3", "v3")
    assert ai_summarize._read_cache(tmp_path, "k3") == "v3"


def test_summarize_issue_mock_and_default_paths(monkeypatch) -> None:
    issue = {"id": "kanbus-6", "title": "Title"}
    monkeypatch.setenv("KANBUS_TEST_AI_MOCK", "1")
    assert (
        ai_summarize.summarize_issue(issue, "short", _ai_config())
        == "Generated summary for kanbus-6"
    )
    monkeypatch.delenv("KANBUS_TEST_AI_MOCK", raising=False)
    assert (
        ai_summarize.summarize_issue(issue, "short", _ai_config()) == "Summary: Title"
    )


def test_make_ai_summarize_supports_issue_dict_input_and_no_cache(
    monkeypatch,
) -> None:
    monkeypatch.delenv("KANBUS_TEST_AI_MOCK", raising=False)
    issue = {"identifier": "kanbus-7", "title": "From dict", "updated_at": "2026-03-09"}
    fn = ai_summarize.make_ai_summarize({}, _ai_config(), None)
    assert fn(issue, "short") == "Summary: From dict"


def test_make_ai_summarize_returns_not_configured_when_ai_missing(
    tmp_path: Path,
) -> None:
    issue = {"identifier": "kanbus-8", "title": "No AI", "updated_at": "2026-03-09"}
    fn = ai_summarize.make_ai_summarize({"kanbus-8": issue}, None, tmp_path)
    assert fn("kanbus-8") == "(AI summarization not configured)"



"""Built-in policy step definitions."""

from __future__ import annotations

import re
from dataclasses import dataclass
from enum import Enum
from typing import Callable, Optional

from kanbus.policy_context import PolicyContext, PolicyOperation


class StepOutcome(str, Enum):
    """Outcome of evaluating a step."""

    PASS = "pass"
    SKIP = "skip"
    FAIL = "fail"
    WARN = "warn"
    SUGGEST = "suggest"
    EXPLAIN = "explain"


StepResult = tuple[StepOutcome, Optional[str]]


class StepCategory(str, Enum):
    """Category of a step."""

    GIVEN = "Given"
    WHEN = "When"
    THEN = "Then"


@dataclass
class StepDefinition:
    """A single step definition with pattern and handler.

    :param description: Human-readable description of what this step does.
    :type description: str
    :param category: Step category (Given/When/Then).
    :type category: StepCategory
    :param usage_pattern: Human-readable usage pattern.
    :type usage_pattern: str
    :param pattern: Regex pattern to match step text.
    :type pattern: re.Pattern
    :param handler: Handler function that evaluates the step.
    :type handler: Callable[[PolicyContext, re.Match], StepResult]
    """

    description: str
    category: StepCategory
    usage_pattern: str
    pattern: re.Pattern
    handler: Callable[[PolicyContext, re.Match], StepResult]

    def matches(self, text: str) -> Optional[re.Match]:
        """Check if this step matches the given text.

        :param text: Step text to match.
        :type text: str
        :return: Match object if pattern matches.
        :rtype: Optional[re.Match]
        """
        return self.pattern.match(text)

    def execute(self, context: PolicyContext, match: re.Match) -> StepResult:
        """Execute the step handler.

        :param context: Policy evaluation context.
        :type context: PolicyContext
        :param match: Regex match object with captured groups.
        :type match: re.Match
        :return: Step outcome and optional error message.
        :rtype: StepResult
        """
        return self.handler(context, match)


class StepRegistry:
    """Registry of all built-in step definitions."""

    def __init__(self) -> None:
        """Initialize registry with all built-in steps."""
        self.steps = build_step_definitions()

    def find_step(self, text: str) -> Optional[tuple[StepDefinition, re.Match]]:
        """Find a step definition matching the given text.

        :param text: Step text to match.
        :type text: str
        :return: Tuple of step definition and match object if found.
        :rtype: Optional[tuple[StepDefinition, re.Match]]
        """
        for step in self.steps:
            match = step.matches(text)
            if match:
                return (step, match)
        return None


def build_step_definitions() -> list[StepDefinition]:
    """Build the list of all built-in step definitions.

    :return: List of step definitions.
    :rtype: list[StepDefinition]
    """
    return [
        StepDefinition(
            "Filter by issue type",
            StepCategory.GIVEN,
            'the issue type is "TYPE"',
            re.compile(r'^the issue type is "([^"]+)"$'),
            given_issue_type_is,
        ),
        StepDefinition(
            "Filter by label presence",
            StepCategory.GIVEN,
            'the issue has label "LABEL"',
            re.compile(r'^the issue has label "([^"]+)"$'),
            given_issue_has_label,
        ),
        StepDefinition(
            "Filter by parent presence",
            StepCategory.GIVEN,
            "the issue has a parent",
            re.compile(r"^the issue has a parent$"),
            given_issue_has_parent,
        ),
        StepDefinition(
            "Filter by priority",
            StepCategory.GIVEN,
            "the issue priority is N",
            re.compile(r"^the issue priority is (\d+)$"),
            given_issue_priority_is,
        ),
        StepDefinition(
            "Filter by issue status",
            StepCategory.GIVEN,
            'the issue status is "STATUS"',
            re.compile(r'^the issue status is "([^"]+)"$'),
            given_issue_status_is,
        ),
        StepDefinition(
            "Filter by transition target",
            StepCategory.WHEN,
            'transitioning to "STATUS"',
            re.compile(r'^transitioning to "([^"]+)"$'),
            when_transitioning_to,
        ),
        StepDefinition(
            "Filter by transition source",
            StepCategory.WHEN,
            'transitioning from "STATUS"',
            re.compile(r'^transitioning from "([^"]+)"$'),
            when_transitioning_from,
        ),
        StepDefinition(
            "Filter by specific transition",
            StepCategory.WHEN,
            'transitioning from "A" to "B"',
            re.compile(r'^transitioning from "([^"]+)" to "([^"]+)"$'),
            when_transitioning_from_to,
        ),
        StepDefinition(
            "Filter by create operation",
            StepCategory.WHEN,
            "creating an issue",
            re.compile(r"^creating an issue$"),
            when_creating_issue,
        ),
        StepDefinition(
            "Filter by close operation",
            StepCategory.WHEN,
            "closing an issue",
            re.compile(r"^closing an issue$"),
            when_closing_issue,
        ),
        StepDefinition(
            "Filter by update operation",
            StepCategory.WHEN,
            "updating an issue",
            re.compile(r"^updating an issue$"),
            when_updating_issue,
        ),
        StepDefinition(
            "Filter by delete operation",
            StepCategory.WHEN,
            "deleting an issue",
            re.compile(r"^deleting an issue$"),
            when_deleting_issue,
        ),
        StepDefinition(
            "Filter by view operation",
            StepCategory.WHEN,
            "viewing an issue",
            re.compile(r"^viewing an issue$"),
            when_viewing_issue,
        ),
        StepDefinition(
            "Filter by list operation",
            StepCategory.WHEN,
            "listing issues",
            re.compile(r"^listing issues$"),
            when_listing_issues,
        ),
        StepDefinition(
            "Filter by ready-list operation",
            StepCategory.WHEN,
            "listing ready issues",
            re.compile(r"^listing ready issues$"),
            when_listing_ready_issues,
        ),
        StepDefinition(
            "Assert field is set",
            StepCategory.THEN,
            'the issue must have field "FIELD"',
            re.compile(r'^the issue must have field "([^"]+)"$'),
            then_issue_must_have_field,
        ),
        StepDefinition(
            "Assert field is not set",
            StepCategory.THEN,
            'the issue must not have field "FIELD"',
            re.compile(r'^the issue must not have field "([^"]+)"$'),
            then_issue_must_not_have_field,
        ),
        StepDefinition(
            "Assert field equals value",
            StepCategory.THEN,
            'the field "FIELD" must be "VALUE"',
            re.compile(r'^the field "([^"]+)" must be "([^"]+)"$'),
            then_field_must_be,
        ),
        StepDefinition(
            "Assert all children have status",
            StepCategory.THEN,
            'all child issues must have status "STATUS"',
            re.compile(r'^all child issues must have status "([^"]+)"$'),
            then_all_children_must_have_status,
        ),
        StepDefinition(
            "Assert no children have status",
            StepCategory.THEN,
            'no child issues may have status "STATUS"',
            re.compile(r'^no child issues may have status "([^"]+)"$'),
            then_no_children_may_have_status,
        ),
        StepDefinition(
            "Assert minimum child count",
            StepCategory.THEN,
            "the issue must have at least N child issues",
            re.compile(r"^the issue must have at least (\d+) child issues?$"),
            then_issue_must_have_at_least_n_child_issues,
        ),
        StepDefinition(
            "Assert parent has status",
            StepCategory.THEN,
            'the parent issue must have status "STATUS"',
            re.compile(r'^the parent issue must have status "([^"]+)"$'),
            then_parent_must_have_status,
        ),
        StepDefinition(
            "Assert minimum label count",
            StepCategory.THEN,
            "the issue must have at least N labels",
            re.compile(r"^the issue must have at least (\d+) labels?$"),
            then_issue_must_have_at_least_n_labels,
        ),
        StepDefinition(
            "Assert has specific label",
            StepCategory.THEN,
            'the issue must have label "LABEL"',
            re.compile(r'^the issue must have label "([^"]+)"$'),
            then_issue_must_have_label,
        ),
        StepDefinition(
            "Assert description not empty",
            StepCategory.THEN,
            "the description must not be empty",
            re.compile(r"^the description must not be empty$"),
            then_description_must_not_be_empty,
        ),
        StepDefinition(
            "Assert title matches pattern",
            StepCategory.THEN,
            'the title must match pattern "REGEX"',
            re.compile(r'^the title must match pattern "([^"]+)"$'),
            then_title_must_match_pattern,
        ),
        StepDefinition(
            "Filter by custom field presence",
            StepCategory.GIVEN,
            'the custom field "FIELD" is set',
            re.compile(r'^the custom field "([^"]+)" is set$'),
            given_custom_field_is_set,
        ),
        StepDefinition(
            "Assert custom field is set",
            StepCategory.THEN,
            'the custom field "FIELD" must be set',
            re.compile(r'^the custom field "([^"]+)" must be set$'),
            then_custom_field_must_be_set,
        ),
        StepDefinition(
            "Assert custom field equals value",
            StepCategory.THEN,
            'the custom field "FIELD" must be "VALUE"',
            re.compile(r'^the custom field "([^"]+)" must be "([^"]+)"$'),
            then_custom_field_must_be,
        ),
        StepDefinition(
            "Emit warning guidance",
            StepCategory.THEN,
            'warn "TEXT"',
            re.compile(r'^warn "([^"]*)"$'),
            then_warn,
        ),
        StepDefinition(
            "Emit suggestion guidance",
            StepCategory.THEN,
            'suggest "TEXT"',
            re.compile(r'^suggest "([^"]*)"$'),
            then_suggest,
        ),
        StepDefinition(
            "Attach explanation to previous item",
            StepCategory.THEN,
            'explain "TEXT"',
            re.compile(r'^explain "([^"]*)"$'),
            then_explain,
        ),
    ]


def given_issue_type_is(context: PolicyContext, match: re.Match) -> StepResult:
    """Filter by issue type."""
    expected_type = match.group(1)
    if context.issue.issue_type == expected_type:
        return (StepOutcome.PASS, None)
    return (StepOutcome.SKIP, None)


def given_issue_has_label(context: PolicyContext, match: re.Match) -> StepResult:
    """Filter by label presence."""
    label = match.group(1)
    if label in context.issue.labels:
        return (StepOutcome.PASS, None)
    return (StepOutcome.SKIP, None)


def given_issue_has_parent(context: PolicyContext, match: re.Match) -> StepResult:
    """Filter by parent presence."""
    if context.issue.parent:
        return (StepOutcome.PASS, None)
    return (StepOutcome.SKIP, None)


def given_issue_priority_is(context: PolicyContext, match: re.Match) -> StepResult:
    """Filter by priority."""
    priority = int(match.group(1))
    if context.issue.priority == priority:
        return (StepOutcome.PASS, None)
    return (StepOutcome.SKIP, None)


def given_issue_status_is(context: PolicyContext, match: re.Match) -> StepResult:
    """Filter by issue status."""
    expected_status = match.group(1)
    if context.issue.status == expected_status:
        return (StepOutcome.PASS, None)
    return (StepOutcome.SKIP, None)


def when_transitioning_to(context: PolicyContext, match: re.Match) -> StepResult:
    """Filter by transition target."""
    status = match.group(1)
    if context.is_transitioning_to(status):
        return (StepOutcome.PASS, None)
    return (StepOutcome.SKIP, None)


def when_transitioning_from(context: PolicyContext, match: re.Match) -> StepResult:
    """Filter by transition source."""
    status = match.group(1)
    if context.is_transitioning_from(status):
        return (StepOutcome.PASS, None)
    return (StepOutcome.SKIP, None)


def when_transitioning_from_to(context: PolicyContext, match: re.Match) -> StepResult:
    """Filter by specific transition."""
    from_status = match.group(1)
    to_status = match.group(2)
    if context.is_transitioning_from(from_status) and context.is_transitioning_to(
        to_status
    ):
        return (StepOutcome.PASS, None)
    return (StepOutcome.SKIP, None)


def when_creating_issue(context: PolicyContext, match: re.Match) -> StepResult:
    """Filter by create operation."""
    if context.operation == PolicyOperation.CREATE:
        return (StepOutcome.PASS, None)
    return (StepOutcome.SKIP, None)


def when_closing_issue(context: PolicyContext, match: re.Match) -> StepResult:
    """Filter by close operation."""
    if context.operation == PolicyOperation.CLOSE or (
        context.operation == PolicyOperation.UPDATE
        and context.is_transitioning_to("closed")
    ):
        return (StepOutcome.PASS, None)
    return (StepOutcome.SKIP, None)


def when_updating_issue(context: PolicyContext, match: re.Match) -> StepResult:
    """Filter by update operation."""
    if context.operation == PolicyOperation.UPDATE:
        return (StepOutcome.PASS, None)
    return (StepOutcome.SKIP, None)


def when_deleting_issue(context: PolicyContext, match: re.Match) -> StepResult:
    """Filter by delete operation."""
    if context.operation == PolicyOperation.DELETE:
        return (StepOutcome.PASS, None)
    return (StepOutcome.SKIP, None)


def when_viewing_issue(context: PolicyContext, match: re.Match) -> StepResult:
    """Filter by view operation."""
    if context.operation == PolicyOperation.VIEW:
        return (StepOutcome.PASS, None)
    return (StepOutcome.SKIP, None)


def when_listing_issues(context: PolicyContext, match: re.Match) -> StepResult:
    """Filter by list operation."""
    if context.operation == PolicyOperation.LIST:
        return (StepOutcome.PASS, None)
    return (StepOutcome.SKIP, None)


def when_listing_ready_issues(context: PolicyContext, match: re.Match) -> StepResult:
    """Filter by ready-list operation."""
    if context.operation == PolicyOperation.READY:
        return (StepOutcome.PASS, None)
    return (StepOutcome.SKIP, None)


def then_issue_must_have_field(context: PolicyContext, match: re.Match) -> StepResult:
    """Assert field is set."""
    field = match.group(1)
    issue = context.issue

    field_map = {
        "assignee": issue.assignee is not None,
        "parent": issue.parent is not None,
        "description": bool(issue.description.strip()),
        "title": bool(issue.title.strip()),
        "creator": issue.creator is not None,
    }

    if field not in field_map:
        return (StepOutcome.FAIL, f"unknown field: {field}")

    if field_map[field]:
        return (StepOutcome.PASS, None)
    return (StepOutcome.FAIL, f'issue does not have field "{field}" set')


def then_issue_must_not_have_field(
    context: PolicyContext, match: re.Match
) -> StepResult:
    """Assert field is not set."""
    field = match.group(1)
    issue = context.issue

    field_map = {
        "assignee": issue.assignee is not None,
        "parent": issue.parent is not None,
        "creator": issue.creator is not None,
    }

    if field not in field_map:
        return (StepOutcome.FAIL, f"unknown field: {field}")

    if not field_map[field]:
        return (StepOutcome.PASS, None)
    return (StepOutcome.FAIL, f'issue has field "{field}" set but should not')


def then_field_must_be(context: PolicyContext, match: re.Match) -> StepResult:
    """Assert field equals value."""
    field = match.group(1)
    expected_value = match.group(2)
    issue = context.issue

    field_map = {
        "status": issue.status,
        "issue_type": issue.issue_type,
        "type": issue.issue_type,
        "assignee": issue.assignee,
        "parent": issue.parent,
        "creator": issue.creator,
    }

    if field not in field_map:
        return (StepOutcome.FAIL, f"unknown field: {field}")

    actual_value = field_map[field]
    if actual_value == expected_value:
        return (StepOutcome.PASS, None)
    if actual_value is None:
        return (StepOutcome.FAIL, f'field "{field}" is not set')
    return (
        StepOutcome.FAIL,
        f'field "{field}" is "{actual_value}" but must be "{expected_value}"',
    )


def then_all_children_must_have_status(
    context: PolicyContext, match: re.Match
) -> StepResult:
    """Assert all children have status."""
    required_status = match.group(1)
    children = context.child_issues()

    if not children:
        return (StepOutcome.PASS, None)

    non_matching = [c for c in children if c.status != required_status]

    if not non_matching:
        return (StepOutcome.PASS, None)

    ids = ", ".join(c.identifier for c in non_matching)
    return (
        StepOutcome.FAIL,
        f'child issues {ids} do not have status "{required_status}"',
    )


def then_no_children_may_have_status(
    context: PolicyContext, match: re.Match
) -> StepResult:
    """Assert no children have status."""
    forbidden_status = match.group(1)
    children = context.child_issues()

    matching = [c for c in children if c.status == forbidden_status]

    if not matching:
        return (StepOutcome.PASS, None)

    ids = ", ".join(c.identifier for c in matching)
    return (
        StepOutcome.FAIL,
        f'child issues {ids} have status "{forbidden_status}" but should not',
    )


def then_issue_must_have_at_least_n_child_issues(
    context: PolicyContext, match: re.Match
) -> StepResult:
    """Assert minimum child issue count."""
    min_count = int(match.group(1))
    actual_count = len(context.child_issues())
    if actual_count >= min_count:
        return (StepOutcome.PASS, None)
    return (
        StepOutcome.FAIL,
        f"issue has {actual_count} child issue(s) but must have at least {min_count}",
    )


def then_parent_must_have_status(context: PolicyContext, match: re.Match) -> StepResult:
    """Assert parent has status."""
    required_status = match.group(1)
    parent = context.parent_issue()

    if parent is None:
        return (StepOutcome.FAIL, "issue has no parent")

    if parent.status == required_status:
        return (StepOutcome.PASS, None)

    return (
        StepOutcome.FAIL,
        f'parent issue {parent.identifier} has status "{parent.status}" '
        f'but must have status "{required_status}"',
    )


def then_issue_must_have_at_least_n_labels(
    context: PolicyContext, match: re.Match
) -> StepResult:
    """Assert minimum label count."""
    min_count = int(match.group(1))
    actual_count = len(context.issue.labels)

    if actual_count >= min_count:
        return (StepOutcome.PASS, None)

    return (
        StepOutcome.FAIL,
        f"issue has {actual_count} label(s) but must have at least {min_count}",
    )


def then_issue_must_have_label(context: PolicyContext, match: re.Match) -> StepResult:
    """Assert has specific label."""
    required_label = match.group(1)
    if required_label in context.issue.labels:
        return (StepOutcome.PASS, None)
    return (StepOutcome.FAIL, f'issue does not have label "{required_label}"')


def then_description_must_not_be_empty(
    context: PolicyContext, match: re.Match
) -> StepResult:
    """Assert description not empty."""
    if context.issue.description.strip():
        return (StepOutcome.PASS, None)
    return (StepOutcome.FAIL, "issue description is empty")


def then_title_must_match_pattern(
    context: PolicyContext, match: re.Match
) -> StepResult:
    """Assert title matches pattern."""
    pattern_str = match.group(1)
    try:
        pattern = re.compile(pattern_str)
    except re.error as error:
        return (StepOutcome.FAIL, f"invalid regex pattern: {error}")

    if pattern.search(context.issue.title):
        return (StepOutcome.PASS, None)

    return (
        StepOutcome.FAIL,
        f'title "{context.issue.title}" does not match pattern "{pattern_str}"',
    )


def given_custom_field_is_set(context: PolicyContext, match: re.Match) -> StepResult:
    """Filter by custom field presence."""
    field = match.group(1)
    if field in context.issue.custom:
        return (StepOutcome.PASS, None)
    return (StepOutcome.SKIP, None)


def then_custom_field_must_be_set(
    context: PolicyContext, match: re.Match
) -> StepResult:
    """Assert custom field is set."""
    field = match.group(1)
    if field in context.issue.custom:
        return (StepOutcome.PASS, None)
    return (StepOutcome.FAIL, f'custom field "{field}" is not set')


def then_custom_field_must_be(context: PolicyContext, match: re.Match) -> StepResult:
    """Assert custom field equals value."""
    field = match.group(1)
    expected_value = match.group(2)

    if field not in context.issue.custom:
        return (StepOutcome.FAIL, f'custom field "{field}" is not set')

    actual_value = str(context.issue.custom[field])
    if actual_value == expected_value:
        return (StepOutcome.PASS, None)

    return (
        StepOutcome.FAIL,
        f'custom field "{field}" is "{actual_value}" but must be "{expected_value}"',
    )


def then_warn(context: PolicyContext, match: re.Match) -> StepResult:
    """Emit non-blocking warning guidance."""
    return (StepOutcome.WARN, match.group(1))


def then_suggest(context: PolicyContext, match: re.Match) -> StepResult:
    """Emit non-blocking suggestion guidance."""
    return (StepOutcome.SUGGEST, match.group(1))


def then_explain(context: PolicyContext, match: re.Match) -> StepResult:
    """Attach explanation text to the previously emitted item."""
    return (StepOutcome.EXPLAIN, match.group(1))



from __future__ import annotations

from pathlib import Path

import pytest

from kanbus import policy_guidance
from kanbus.policy_context import (
    GuidanceItem,
    GuidanceSeverity,
    PolicyEvaluationReport,
    PolicyOperation,
    PolicyViolationError,
)

from test_helpers import build_issue


def test_guidance_is_enabled_honors_flag_and_env(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    assert policy_guidance.guidance_is_enabled(no_guidance=True) is False

    monkeypatch.setenv("KANBUS_NO_GUIDANCE", "1")
    assert policy_guidance.guidance_is_enabled() is False

    monkeypatch.setenv("KANBUS_NO_GUIDANCE", "TRUE")
    assert policy_guidance.guidance_is_enabled() is False

    monkeypatch.setenv("KANBUS_NO_GUIDANCE", "off")
    assert policy_guidance.guidance_is_enabled() is True


def test_collect_guidance_for_issue_returns_empty_without_policies_dir(
    tmp_path: Path,
) -> None:
    issue = build_issue("kanbus-1")
    monkeypatch = pytest.MonkeyPatch()
    monkeypatch.setattr(
        policy_guidance, "load_project_directory", lambda _root: tmp_path
    )
    report = policy_guidance.collect_guidance_for_issue(
        tmp_path, issue, PolicyOperation.UPDATE
    )
    monkeypatch.undo()
    assert report.violations == []
    assert report.guidance_items == []


def test_collect_guidance_for_issue_returns_empty_with_no_policy_documents(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    project_dir = tmp_path / "project"
    (project_dir / "policies").mkdir(parents=True)

    monkeypatch.setattr(
        policy_guidance, "load_project_directory", lambda _root: project_dir
    )
    monkeypatch.setattr(policy_guidance, "load_policies", lambda _policies_dir: [])

    issue = build_issue("kanbus-1")
    report = policy_guidance.collect_guidance_for_issue(
        tmp_path, issue, PolicyOperation.UPDATE
    )
    assert report.violations == []
    assert report.guidance_items == []


def test_collect_guidance_for_issue_builds_context_and_uses_guidance_mode(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    issue = build_issue("kanbus-1")
    all_issues = [issue, build_issue("kanbus-2")]
    project_dir = tmp_path / "project"
    policies_dir = project_dir / "policies"
    policies_dir.mkdir(parents=True)

    seen: dict[str, object] = {}
    documents = [("rules.feature", object())]

    monkeypatch.setattr(
        policy_guidance, "load_project_directory", lambda _root: project_dir
    )
    monkeypatch.setattr(policy_guidance, "load_policies", lambda _d: documents)
    monkeypatch.setattr(
        policy_guidance,
        "get_configuration_path",
        lambda _d: project_dir / "config.yaml",
    )
    monkeypatch.setattr(
        policy_guidance, "load_project_configuration", lambda _p: {"ok": True}
    )
    monkeypatch.setattr(
        policy_guidance, "load_issues_from_directory", lambda _d: all_issues
    )

    def fake_evaluate(
        context: object, docs: object, options: object
    ) -> PolicyEvaluationReport:
        seen["context"] = context
        seen["docs"] = docs
        seen["options"] = options
        return PolicyEvaluationReport(
            guidance_items=[
                GuidanceItem(
                    severity=GuidanceSeverity.SUGGESTION,
                    message="msg",
                    explanations=[],
                    policy_file="rules.feature",
                    scenario="Scenario",
                    step="Then suggest",
                )
            ]
        )

    monkeypatch.setattr(policy_guidance, "evaluate_policies_report", fake_evaluate)

    report = policy_guidance.collect_guidance_for_issue(
        tmp_path, issue, PolicyOperation.CREATE
    )

    assert len(report.guidance_items) == 1
    assert seen["docs"] == documents
    options = seen["options"]
    assert options.collect_all_violations is True
    assert str(options.mode).endswith("GUIDANCE")
    context = seen["context"]
    assert context.current_issue.identifier == "kanbus-1"
    assert context.proposed_issue.identifier == "kanbus-1"
    assert context.transition is None
    assert context.operation == PolicyOperation.CREATE
    assert context.all_issues == all_issues


def test_emit_guidance_for_issues_skips_when_disabled(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    calls: list[str] = []
    monkeypatch.setattr(
        policy_guidance, "guidance_is_enabled", lambda _no_guidance: False
    )
    monkeypatch.setattr(
        policy_guidance,
        "collect_guidance_for_issue",
        lambda *_args, **_kwargs: calls.append("collect"),
    )

    policy_guidance.emit_guidance_for_issues(
        Path("/repo"), [build_issue("kanbus-1")], PolicyOperation.UPDATE
    )
    assert calls == []


def test_emit_guidance_for_issues_emits_sorted_deduped_items_and_converts_violations(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    issue = build_issue("kanbus-1")
    duplicate = GuidanceItem(
        severity=GuidanceSeverity.SUGGESTION,
        message="consider tags",
        explanations=["add labels"],
        policy_file="b.policy",
        scenario="S2",
        step="Then suggest",
    )
    warning = GuidanceItem(
        severity=GuidanceSeverity.WARNING,
        message="missing owner",
        explanations=["Explanation: owner helps triage", "set assignee"],
        policy_file="a.policy",
        scenario="S1",
        step="Then warn",
    )
    violation = PolicyViolationError(
        policy_file="c.policy",
        scenario="S3",
        failed_step="Then ...",
        message="bad policy",
        issue_id="kanbus-1",
    )
    report = PolicyEvaluationReport(
        guidance_items=[duplicate, duplicate, warning],
        violations=[violation],
    )

    monkeypatch.setattr(
        policy_guidance, "guidance_is_enabled", lambda _no_guidance: True
    )
    monkeypatch.setattr(
        policy_guidance,
        "collect_guidance_for_issue",
        lambda *_args, **_kwargs: report,
    )

    lines: list[str] = []
    monkeypatch.setattr(
        policy_guidance.click, "echo", lambda msg, err=True: lines.append(msg)
    )

    policy_guidance.emit_guidance_for_issues(
        Path("/repo"), [issue], PolicyOperation.UPDATE
    )

    assert sum("GUIDANCE WARNING: missing owner" in line for line in lines) == 1
    assert sum("GUIDANCE SUGGESTION: consider tags" in line for line in lines) == 1
    assert any(
        "Guidance policy error (c.policy / Rule: S3): bad policy" in line
        for line in lines
    )
    assert any("  Explanation: owner helps triage" == line for line in lines)
    assert any("  Explanation: set assignee" == line for line in lines)


def test_emit_guidance_for_issues_ignores_collection_exceptions(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    issue = build_issue("kanbus-1")
    monkeypatch.setattr(
        policy_guidance, "guidance_is_enabled", lambda _no_guidance: True
    )

    def _boom(*_args, **_kwargs):
        raise RuntimeError("broken")

    monkeypatch.setattr(policy_guidance, "collect_guidance_for_issue", _boom)
    monkeypatch.setattr(
        policy_guidance.click,
        "echo",
        lambda *_args, **_kwargs: (_ for _ in ()).throw(
            AssertionError("should not echo")
        ),
    )

    policy_guidance.emit_guidance_for_issues(
        Path("/repo"), [issue], PolicyOperation.UPDATE
    )


def test_sorted_deduped_guidance_items_orders_warning_before_suggestion() -> None:
    warning = GuidanceItem(
        severity=GuidanceSeverity.WARNING,
        message="warn",
        explanations=["w"],
        policy_file="z.policy",
        scenario="Z",
        step="Then warn",
    )
    suggestion = GuidanceItem(
        severity=GuidanceSeverity.SUGGESTION,
        message="suggest",
        explanations=["s"],
        policy_file="a.policy",
        scenario="A",
        step="Then suggest",
    )

    result = policy_guidance.sorted_deduped_guidance_items(
        [suggestion, warning, warning]
    )

    assert len(result) == 2
    assert result[0].severity == GuidanceSeverity.WARNING
    assert result[1].severity == GuidanceSeverity.SUGGESTION



from kanbus.github_security_sync import (
    _extract_repo_slug,
    _parse_next_link,
    _severity_to_priority,
)


def test_extract_repo_slug() -> None:
    assert (
        _extract_repo_slug("https://github.com/AnthusAI/Kanbus.git")
        == "AnthusAI/Kanbus"
    )
    assert _extract_repo_slug("git@github.com:AnthusAI/Kanbus.git") == "AnthusAI/Kanbus"
    assert _extract_repo_slug("ssh://gitlab.com/foo/bar.git") is None


def test_parse_next_link() -> None:
    header = (
        '<https://api.github.com/foo?page=2>; rel="next", '
        '<https://api.github.com/foo?page=3>; rel="last"'
    )
    assert _parse_next_link(header) == "https://api.github.com/foo?page=2"
    assert _parse_next_link('<https://api.github.com/foo?page=3>; rel="last"') is None


def test_severity_to_priority() -> None:
    assert _severity_to_priority("critical") == 0
    assert _severity_to_priority("high") == 1
    assert _severity_to_priority("medium") == 2
    assert _severity_to_priority("low") == 3
    assert _severity_to_priority("unknown") == 3



from __future__ import annotations

from pathlib import Path

from kanbus.config_loader import load_project_configuration


def _write_minimal_config(path: Path) -> None:
    path.write_text("project_key: kanbus\n", encoding="utf-8")


def test_load_configuration_accepts_mqtt_override_fields(tmp_path: Path) -> None:
    config_path = tmp_path / ".kanbus.yml"
    _write_minimal_config(config_path)
    (tmp_path / ".kanbus.override.yml").write_text(
        "\n".join(
            [
                "realtime:",
                "  mqtt_custom_authorizer_name: token-auth",
                "  mqtt_api_token: token-value",
            ]
        )
        + "\n",
        encoding="utf-8",
    )

    configuration = load_project_configuration(config_path)

    assert configuration.realtime.mqtt_custom_authorizer_name == "token-auth"
    assert configuration.realtime.mqtt_api_token == "token-value"


def test_load_configuration_applies_mqtt_environment_overrides(
    tmp_path: Path, monkeypatch
) -> None:
    config_path = tmp_path / ".kanbus.yml"
    _write_minimal_config(config_path)
    monkeypatch.setenv("KANBUS_REALTIME_MQTT_CUSTOM_AUTHORIZER_NAME", "env-auth")
    monkeypatch.setenv("KANBUS_REALTIME_MQTT_API_TOKEN", "env-token")

    configuration = load_project_configuration(config_path)

    assert configuration.realtime.mqtt_custom_authorizer_name == "env-auth"
    assert configuration.realtime.mqtt_api_token == "env-token"



from __future__ import annotations

from kanbus.policy_context import (
    GuidanceItem,
    GuidanceSeverity,
    PolicyContext,
    PolicyEvaluationReport,
    PolicyOperation,
    PolicyViolationError,
    StatusTransition,
)

from test_helpers import build_issue, build_project_configuration


def build_context() -> PolicyContext:
    parent = build_issue("kanbus-parent", status="open")
    issue = build_issue("kanbus-child", parent="kanbus-parent", status="in_progress")
    sibling = build_issue("kanbus-sibling")
    child = build_issue("kanbus-sub", parent="kanbus-child")
    return PolicyContext(
        current_issue=issue,
        proposed_issue=issue,
        transition=StatusTransition(from_status="open", to_status="in_progress"),
        operation=PolicyOperation.UPDATE,
        project_configuration=build_project_configuration(),
        all_issues=[parent, issue, sibling, child],
    )


def test_policy_context_transition_and_lookup_helpers() -> None:
    context = build_context()
    assert context.issue.identifier == "kanbus-child"
    assert context.is_transition() is True
    assert context.is_transitioning_to("in_progress") is True
    assert context.is_transitioning_from("open") is True
    assert context.is_transitioning_to("closed") is False
    assert context.is_transitioning_from("closed") is False
    assert [issue.identifier for issue in context.child_issues()] == ["kanbus-sub"]
    assert context.parent_issue() is not None
    assert context.parent_issue().identifier == "kanbus-parent"  # type: ignore[union-attr]


def test_policy_context_parent_issue_handles_missing_parent() -> None:
    context = build_context()
    context.proposed_issue = context.proposed_issue.model_copy(update={"parent": None})
    assert context.parent_issue() is None

    context.proposed_issue = context.proposed_issue.model_copy(
        update={"parent": "nope"}
    )
    assert context.parent_issue() is None

    context.transition = None
    assert context.is_transition() is False
    assert context.is_transitioning_to("in_progress") is False
    assert context.is_transitioning_from("open") is False


def test_policy_violation_error_formats_explanations_and_sorted_guidance() -> None:
    guidance_items = [
        GuidanceItem(
            severity=GuidanceSeverity.SUGGESTION,
            message="later suggestion",
            explanations=["suggest detail"],
            policy_file="b.policy",
            scenario="B",
            step='suggest "later"',
        ),
        GuidanceItem(
            severity=GuidanceSeverity.WARNING,
            message="first warning",
            explanations=["warn detail"],
            policy_file="a.policy",
            scenario="A",
            step='warn "first"',
        ),
    ]

    error = PolicyViolationError(
        policy_file="root.policy",
        scenario="Create rule",
        failed_step='Then the issue must have field "title"',
        message="title is required",
        issue_id="kanbus-1",
        explanations=["because policy"],
        guidance_items=guidance_items,
    )
    text = str(error)
    assert "policy violation in root.policy for issue kanbus-1" in text
    assert "Explanation: because policy" in text
    # warnings should sort ahead of suggestions
    assert text.index("GUIDANCE WARNING: first warning") < text.index(
        "GUIDANCE SUGGESTION: later suggestion"
    )
    assert "Explanation: warn detail" in text
    assert "Explanation: suggest detail" in text


def test_policy_evaluation_report_defaults_are_empty() -> None:
    report = PolicyEvaluationReport()
    assert report.violations == []
    assert report.guidance_items == []



"""IPC protocol models and version checks."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, Optional, Tuple

from pydantic import BaseModel, Field

PROTOCOL_VERSION = "1.0"


class ProtocolError(RuntimeError):
    """Raised when protocol validation fails."""


class RequestEnvelope(BaseModel):
    """Client request envelope for daemon IPC."""

    protocol_version: str = Field(min_length=1)
    request_id: str = Field(min_length=1)
    action: str = Field(min_length=1)
    payload: Dict[str, Any]


class ErrorEnvelope(BaseModel):
    """Structured error payload for daemon responses."""

    code: str = Field(min_length=1)
    message: str = Field(min_length=1)
    details: Dict[str, Any] = Field(default_factory=dict)


class ResponseEnvelope(BaseModel):
    """Daemon response envelope for IPC."""

    protocol_version: str = Field(min_length=1)
    request_id: str = Field(min_length=1)
    status: str = Field(min_length=1)
    result: Optional[Dict[str, Any]] = None
    error: Optional[ErrorEnvelope] = None


@dataclass(frozen=True)
class ProtocolVersions:
    """Parsed protocol versions."""

    client: Tuple[int, int]
    daemon: Tuple[int, int]


def parse_version(version: str) -> Tuple[int, int]:
    """Parse a protocol version string into a major/minor tuple.

    :param version: Version string in MAJOR.MINOR form.
    :type version: str
    :return: Parsed (major, minor) version.
    :rtype: tuple[int, int]
    :raises ProtocolError: If the version is invalid.
    """
    parts = version.split(".")
    if len(parts) != 2:
        raise ProtocolError("invalid protocol version")
    try:
        major = int(parts[0])
        minor = int(parts[1])
    except ValueError as error:
        raise ProtocolError("invalid protocol version") from error
    return major, minor


def validate_protocol_compatibility(client_version: str, daemon_version: str) -> None:
    """Validate protocol compatibility rules.

    :param client_version: Client protocol version string.
    :type client_version: str
    :param daemon_version: Daemon protocol version string.
    :type daemon_version: str
    :raises ProtocolError: If versions are incompatible.
    """
    client_major, client_minor = parse_version(client_version)
    daemon_major, daemon_minor = parse_version(daemon_version)
    if client_major != daemon_major:
        raise ProtocolError("protocol version mismatch")
    if client_minor > daemon_minor:
        raise ProtocolError("protocol version unsupported")



"""Single-line issue formatting for list output."""

from __future__ import annotations

import os
import sys
from typing import Callable, Dict, Iterable, Optional

import click

from kanbus.ids import format_issue_key
from kanbus.models import IssueData, ProjectConfiguration

STATUS_COLORS = {
    "backlog": "grey",
    "open": "cyan",
    "in_progress": "blue",
    "blocked": "red",
    "closed": "green",
    "deferred": "yellow",
}

PRIORITY_COLORS = {
    0: "red",
    1: "bright_red",
    2: "yellow",
    3: "blue",
    4: "white",
}

# Temporary type color mapping; will be replaced with config-driven values.
TYPE_COLORS = {
    "epic": "magenta",
    "initiative": "bright_magenta",
    "task": "white",
    "sub-task": "white",
    "bug": "red",
    "story": "cyan",
    "chore": "blue",
    "event": "bright_blue",
}

KNOWN_COLORS = {
    "black",
    "red",
    "green",
    "yellow",
    "blue",
    "magenta",
    "cyan",
    "white",
    "bright_black",
    "bright_red",
    "bright_green",
    "bright_yellow",
    "bright_blue",
    "bright_magenta",
    "bright_cyan",
    "bright_white",
}


def _normalize_cli_color(value: str | None) -> str | None:
    if value is None:
        return None
    normalized = value.strip()
    if normalized in ("gray", "grey"):
        normalized = "bright_black"
    return normalized if normalized in KNOWN_COLORS else None


def _resolve_status_color(
    status: str, configuration: ProjectConfiguration | None
) -> str:
    if configuration:
        categories = {
            category.name: category.color for category in configuration.categories
        }
        # Look up color from statuses list
        for status_def in configuration.statuses:
            if status_def.key != status:
                continue
            status_color = _normalize_cli_color(status_def.color)
            if status_color:
                return status_color
            category_color = _normalize_cli_color(
                categories.get(status_def.category or "")
            )
            if category_color:
                return category_color
    return STATUS_COLORS.get(status, "white")


def _resolve_priority_color(
    priority: int, configuration: ProjectConfiguration | None
) -> str:
    if configuration:
        definition = configuration.priorities.get(priority)
        if definition and definition.color:
            return definition.color
    return PRIORITY_COLORS.get(priority, "white")


def _resolve_type_color(
    issue_type: str, configuration: ProjectConfiguration | None
) -> str:
    if configuration and issue_type in configuration.type_colors:
        return configuration.type_colors[issue_type]
    return TYPE_COLORS.get(issue_type, "white")


def _normalize_color(color: str | None) -> str | None:
    return _normalize_cli_color(color)


def _safe_color(colorizer: Callable[[str, str], str], text: str, fg: str | None) -> str:
    normalized = _normalize_color(fg)
    if normalized is None:
        return text
    return colorizer(text, fg=normalized)


def format_issue_line(
    issue: IssueData,
    *,
    porcelain: bool = False,
    colorizer: Callable[[str, str], str] | None = None,
    widths: Dict[str, int] | None = None,
    project_context: bool = False,
    configuration: ProjectConfiguration | None = None,
    use_color: Optional[bool] = None,
) -> str:
    """Render a single-line summary similar to Beads.

    :param issue: Issue to format.
    :type issue: IssueData
    :param porcelain: Disable ANSI color when True.
    :type porcelain: bool
    :param colorizer: Optional function to apply color; defaults to click.style.
    :type colorizer: Callable[[str, str], str] | None
    :param widths: Optional column widths for aligned output.
    :type widths: Dict[str, int] | None
    :param project_context: Whether identifiers should omit the project key.
    :type project_context: bool
    :param configuration: Optional project configuration for color overrides.
    :type configuration: ProjectConfiguration | None
    :param use_color: Force color on/off; when None, use NO_COLOR and TTY.
    :type use_color: Optional[bool]
    :return: Formatted line.
    :rtype: str
    """
    if use_color is None:
        use_color = os.getenv("NO_COLOR") is None and sys.stdout.isatty()
    use_color = use_color and not porcelain
    color = colorizer or click.style
    if not use_color:

        def no_color(text: str, **_kwargs: object) -> str:
            return text

        color = no_color
    priority_color = _resolve_priority_color(issue.priority, configuration)
    status_color = _resolve_status_color(issue.status, configuration)

    formatted_identifier = format_issue_key(
        issue.identifier, project_context=project_context
    )

    parent_value = issue.parent or "-"
    parent_display = (
        format_issue_key(parent_value, project_context=project_context)
        if parent_value != "-"
        else parent_value
    )

    type_display = issue.issue_type[:1].upper()

    if porcelain:
        parts = [
            type_display,
            formatted_identifier,
            parent_display,
            issue.status,
            f"P{issue.priority}",
            issue.title,
        ]
        return " | ".join(parts)

    widths = widths or compute_widths([issue], project_context=project_context)

    type_color = _resolve_type_color(issue.issue_type, configuration)
    type_part = _safe_color(color, type_display.ljust(widths["type"]), type_color)

    parent_value = issue.parent or "-"
    parent_display = (
        format_issue_key(parent_value, project_context=project_context)
        if parent_value != "-"
        else parent_value
    )
    status_part = _safe_color(color, issue.status.ljust(widths["status"]), status_color)
    priority_plain = f"P{issue.priority}".ljust(widths["priority"])
    priority_part = _safe_color(color, priority_plain, priority_color)

    identifier_part = formatted_identifier.ljust(widths["identifier"])
    parent_plain = parent_display.ljust(widths["parent"])
    if parent_value == "-" and use_color:
        parent_part = _safe_color(color, parent_plain, "bright_black")
    else:
        parent_part = parent_plain
    title = issue.title
    prefix = issue.custom.get("project_path")
    prefix_part = f"{prefix} " if prefix else ""

    return (
        f"{prefix_part}"
        f"{type_part} "
        f"{identifier_part} "
        f"{parent_part} "
        f"{status_part} "
        f"{priority_part} "
        f"{title}"
    )


def compute_widths(
    issues: Iterable[IssueData], project_context: bool = False
) -> Dict[str, int]:
    """Compute printable column widths for aligned normal-mode output."""

    status_w = 1
    priority_w = 0
    type_w = 0
    identifier_w = 0
    parent_w = 0

    for issue in issues:
        status_w = max(status_w, len(issue.status))
        priority_w = max(priority_w, len(f"P{issue.priority}"))
        type_w = max(type_w, len(issue.issue_type[:1].upper()))
        formatted_identifier = format_issue_key(
            issue.identifier, project_context=project_context
        )
        identifier_w = max(identifier_w, len(formatted_identifier))
        parent_value = issue.parent or "-"
        parent_display = (
            format_issue_key(parent_value, project_context=project_context)
            if parent_value != "-"
            else parent_value
        )
        parent_w = max(parent_w, len(parent_display))

    return {
        "status": status_w,
        "priority": priority_w,
        "type": type_w,
        "identifier": identifier_w,
        "parent": parent_w,
    }



"""File edit operations mirroring the Anthropic text editor tool.

Provides view, str_replace, create, and insert commands with strict single-match
replacement semantics for agent compatibility.
"""

from __future__ import annotations

from pathlib import Path


class TextEditorError(RuntimeError):
    """Raised when a text editor operation fails."""


def edit_view(root: Path, path: Path, view_range: tuple[int, int] | None = None) -> str:
    """View file contents or directory listing.

    :param root: Repository root (current working directory for relative paths).
    :type root: Path
    :param path: File or directory path (relative to root).
    :type path: Path
    :param view_range: Optional (start, end) line numbers (1-indexed; -1 for end).
    :type view_range: tuple[int, int] | None
    :return: File contents with optional line numbers, or directory listing.
    :rtype: str
    :raises TextEditorError: If path does not exist or is outside root.
    """
    resolved = (root / path).resolve()
    if not resolved.is_relative_to(root.resolve()):
        raise TextEditorError("path escapes repository root")
    if not resolved.exists():
        raise TextEditorError("file or directory not found")
    if resolved.is_dir():
        entries = sorted(
            resolved.iterdir(), key=lambda p: (not p.is_dir(), p.name.lower())
        )
        return "\n".join(p.name + ("/" if p.is_dir() else "") for p in entries)
    content = resolved.read_text(encoding="utf-8")
    lines = content.splitlines()
    if view_range is not None:
        start_one, end_one = view_range
        start_idx = max(0, start_one - 1) if start_one >= 1 else 0
        end_idx = len(lines) if end_one == -1 else min(len(lines), end_one)
        lines = lines[start_idx:end_idx]
        line_start = start_idx + 1
    else:
        line_start = 1
    return "\n".join(f"{line_start + i}: {line}" for i, line in enumerate(lines))


def edit_str_replace(root: Path, path: Path, old_str: str, new_str: str) -> str:
    """Replace exactly one occurrence of old_str with new_str.

    :param root: Repository root.
    :type root: Path
    :param path: File path (relative to root).
    :type path: Path
    :param old_str: Exact text to replace.
    :type old_str: str
    :param new_str: Replacement text.
    :type new_str: str
    :return: Success message.
    :rtype: str
    :raises TextEditorError: If file not found, 0 matches, or multiple matches.
    """
    resolved = (root / path).resolve()
    if not resolved.is_relative_to(root.resolve()):
        raise TextEditorError("path escapes repository root")
    if not resolved.exists() or not resolved.is_file():
        raise TextEditorError("file not found")
    content = resolved.read_text(encoding="utf-8")
    count = content.count(old_str)
    if count == 0:
        raise TextEditorError("no match found for replacement")
    if count > 1:
        raise TextEditorError(
            f"found {count} matches for replacement text; provide more context for a unique match"
        )
    new_content = content.replace(old_str, new_str, 1)
    resolved.write_text(new_content, encoding="utf-8")
    return "Successfully replaced text at exactly one location."


def edit_create(root: Path, path: Path, file_text: str) -> str:
    """Create a new file with the given content.

    :param root: Repository root.
    :type root: Path
    :param path: File path (relative to root).
    :type path: Path
    :param file_text: Content to write.
    :type file_text: str
    :return: Success message.
    :rtype: str
    :raises TextEditorError: If file already exists or path escapes root.
    """
    resolved = (root / path).resolve()
    if not resolved.is_relative_to(root.resolve()):
        raise TextEditorError("path escapes repository root")
    if resolved.exists():
        raise TextEditorError("file already exists")
    resolved.parent.mkdir(parents=True, exist_ok=True)
    resolved.write_text(file_text, encoding="utf-8")
    return "Successfully created file."


def edit_insert(root: Path, path: Path, insert_line: int, insert_text: str) -> str:
    """Insert text after the given line number.

    insert_line 0 means insert at the beginning. Line numbers are 1-indexed.

    :param root: Repository root.
    :type root: Path
    :param path: File path (relative to root).
    :type path: Path
    :param insert_line: Line number after which to insert (0 = beginning).
    :type insert_line: int
    :param insert_text: Text to insert.
    :type insert_text: str
    :return: Success message.
    :rtype: str
    :raises TextEditorError: If file not found or insert_line is invalid.
    """
    resolved = (root / path).resolve()
    if not resolved.is_relative_to(root.resolve()):
        raise TextEditorError("path escapes repository root")
    if not resolved.exists() or not resolved.is_file():
        raise TextEditorError("file not found")
    lines = resolved.read_text(encoding="utf-8").splitlines()
    if insert_line < 0:
        raise TextEditorError("insert_line must be non-negative")
    if insert_line > len(lines):
        raise TextEditorError("insert_line exceeds file length")
    idx = insert_line
    new_lines = lines[:idx] + insert_text.splitlines() + lines[idx:]
    resolved.write_text(
        "\n".join(new_lines) + ("\n" if new_lines else ""), encoding="utf-8"
    )
    return "Successfully inserted text."



from __future__ import annotations

from pathlib import Path
import runpy
import sys

import pytest

from kanbus import daemon
from kanbus import daemon_server


def test_parse_args_requires_root_and_parses_value() -> None:
    args = daemon.parse_args(["--root", "/tmp/project"])
    assert args.root == "/tmp/project"

    with pytest.raises(SystemExit):
        daemon.parse_args([])


def test_main_delegates_to_run_daemon(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    called: dict[str, Path] = {}
    monkeypatch.setattr(
        daemon, "run_daemon", lambda root: called.setdefault("root", root)
    )
    daemon.main(["--root", str(tmp_path)])
    assert called["root"] == tmp_path


def test_module_main_invocation_path(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    called: dict[str, Path] = {}
    monkeypatch.setattr(
        daemon_server, "run_daemon", lambda root: called.setdefault("root", root)
    )
    monkeypatch.setattr(sys, "argv", ["kanbus.daemon", "--root", str(tmp_path)])
    runpy.run_module("kanbus.daemon", run_name="__main__")
    assert called["root"] == tmp_path



"""Dependency management utilities."""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import List

from kanbus.issue_files import read_issue_from_file, write_issue_to_file
from kanbus.issue_lookup import IssueLookupError, load_issue_from_project
from kanbus.models import DependencyLink, IssueData
from kanbus.event_history import (
    create_event,
    dependency_payload,
    events_dir_for_issue_path,
    now_timestamp,
    write_events_batch,
)
from kanbus.project import (
    ProjectMarkerError,
    discover_project_directories,
    find_project_local_directory,
)
from kanbus.migration import MigrationError, load_beads_issues
from kanbus.users import get_current_user
from kanbus.gossip import publish_issue_mutation

ALLOWED_DEPENDENCY_TYPES = {"blocked-by", "relates-to"}


class DependencyError(RuntimeError):
    """Raised when dependency operations fail."""


@dataclass(frozen=True)
class DependencyGraph:
    """Dependency graph built from blocked-by links."""

    edges: dict[str, list[str]]


def add_dependency(
    root: Path,
    source_id: str,
    target_id: str,
    dependency_type: str,
) -> IssueData:
    """Add a dependency to an issue.

    :param root: Repository root path.
    :type root: Path
    :param source_id: Issue identifier to update.
    :type source_id: str
    :param target_id: Dependency target issue identifier.
    :type target_id: str
    :param dependency_type: Dependency type to add.
    :type dependency_type: str
    :return: Updated issue data.
    :rtype: IssueData
    :raises DependencyError: If the dependency cannot be added.
    """
    _validate_dependency_type(dependency_type)
    try:
        source_lookup = load_issue_from_project(root, source_id)
        load_issue_from_project(root, target_id)
    except (IssueLookupError, ProjectMarkerError) as error:
        raise DependencyError(str(error)) from error

    if dependency_type == "blocked-by":
        _ensure_no_cycle(source_lookup.project_dir, source_id, target_id)

    if _has_dependency(source_lookup.issue, target_id, dependency_type):
        return source_lookup.issue

    updated_issue = source_lookup.issue.model_copy(
        update={
            "dependencies": [
                *source_lookup.issue.dependencies,
                DependencyLink(target=target_id, type=dependency_type),
            ]
        }
    )
    write_issue_to_file(updated_issue, source_lookup.issue_path)
    occurred_at = now_timestamp()
    actor_id = get_current_user()
    event = create_event(
        issue_id=updated_issue.identifier,
        event_type="dependency_added",
        actor_id=actor_id,
        payload=dependency_payload(dependency_type, target_id),
        occurred_at=occurred_at,
    )
    events_dir = events_dir_for_issue_path(
        source_lookup.project_dir, source_lookup.issue_path
    )
    try:
        write_events_batch(events_dir, [event])
    except Exception as error:  # noqa: BLE001
        write_issue_to_file(source_lookup.issue, source_lookup.issue_path)
        raise DependencyError(str(error)) from error
    if source_lookup.issue_path.parent == source_lookup.project_dir / "issues":
        publish_issue_mutation(
            root,
            source_lookup.project_dir,
            updated_issue,
            event.event_id,
            "issue.mutated",
        )
    return updated_issue


def remove_dependency(
    root: Path,
    source_id: str,
    target_id: str,
    dependency_type: str,
) -> IssueData:
    """Remove a dependency from an issue.

    :param root: Repository root path.
    :type root: Path
    :param source_id: Issue identifier to update.
    :type source_id: str
    :param target_id: Dependency target issue identifier.
    :type target_id: str
    :param dependency_type: Dependency type to remove.
    :type dependency_type: str
    :return: Updated issue data.
    :rtype: IssueData
    :raises DependencyError: If the dependency cannot be removed.
    """
    _validate_dependency_type(dependency_type)
    try:
        source_lookup = load_issue_from_project(root, source_id)
    except (IssueLookupError, ProjectMarkerError) as error:
        raise DependencyError(str(error)) from error

    filtered = [
        dependency
        for dependency in source_lookup.issue.dependencies
        if not (
            dependency.target == target_id
            and dependency.dependency_type == dependency_type
        )
    ]
    updated_issue = source_lookup.issue.model_copy(update={"dependencies": filtered})
    write_issue_to_file(updated_issue, source_lookup.issue_path)
    occurred_at = now_timestamp()
    actor_id = get_current_user()
    event = create_event(
        issue_id=updated_issue.identifier,
        event_type="dependency_removed",
        actor_id=actor_id,
        payload=dependency_payload(dependency_type, target_id),
        occurred_at=occurred_at,
    )
    events_dir = events_dir_for_issue_path(
        source_lookup.project_dir, source_lookup.issue_path
    )
    try:
        write_events_batch(events_dir, [event])
    except Exception as error:  # noqa: BLE001
        write_issue_to_file(source_lookup.issue, source_lookup.issue_path)
        raise DependencyError(str(error)) from error
    if source_lookup.issue_path.parent == source_lookup.project_dir / "issues":
        publish_issue_mutation(
            root,
            source_lookup.project_dir,
            updated_issue,
            event.event_id,
            "issue.mutated",
        )
    return updated_issue


def list_ready_issues(
    root: Path,
    include_local: bool = True,
    local_only: bool = False,
    beads_mode: bool = False,
) -> List[IssueData]:
    """List issues that are not blocked by dependencies.

    :param root: Repository root path.
    :type root: Path
    :param beads_mode: Whether to read from Beads JSONL instead of project files.
    :type beads_mode: bool
    :return: Ready issues.
    :rtype: List[IssueData]
    :raises DependencyError: If listing fails.
    """
    if local_only and not include_local:
        raise DependencyError("local-only conflicts with no-local")
    if beads_mode:
        if local_only or not include_local:
            raise DependencyError("beads mode does not support local filtering")
        try:
            issues = load_beads_issues(root)
        except MigrationError as error:
            raise DependencyError(str(error)) from error
        return [
            issue
            for issue in issues
            if issue.status != "closed" and not _blocked_by_dependency(issue)
        ]
    try:
        project_dirs = discover_project_directories(root)
    except ProjectMarkerError as error:
        raise DependencyError(str(error)) from error
    if not project_dirs:
        raise DependencyError("project not initialized")

    issues: List[IssueData] = []
    if len(project_dirs) == 1:
        issues = _load_ready_issues_for_project(
            root, project_dirs[0], include_local, local_only, tag_project=False
        )
    else:
        for project_dir in sorted(project_dirs):
            issues.extend(
                _load_ready_issues_for_project(
                    root, project_dir, include_local, local_only, tag_project=True
                )
            )

    ready = [
        issue
        for issue in issues
        if issue.status != "closed" and not _blocked_by_dependency(issue)
    ]
    return ready


def _load_ready_issues_for_project(
    root: Path,
    project_dir: Path,
    include_local: bool,
    local_only: bool,
    tag_project: bool,
) -> List[IssueData]:
    issues_dir = project_dir / "issues"
    shared_issues = _load_issues_from_directory(issues_dir)
    shared_tagged = [_tag_issue_source(issue, "shared") for issue in shared_issues]
    if tag_project:
        shared_tagged = [
            _tag_issue_project(issue, root, project_dir) for issue in shared_tagged
        ]

    local_tagged: List[IssueData] = []
    if include_local or local_only:
        local_dir = find_project_local_directory(project_dir)
        if local_dir is not None:
            local_issues_dir = local_dir / "issues"
            if local_issues_dir.exists():
                local_tagged = [
                    _tag_issue_source(issue, "local")
                    for issue in _load_issues_from_directory(local_issues_dir)
                ]
                if tag_project:
                    local_tagged = [
                        _tag_issue_project(issue, root, project_dir)
                        for issue in local_tagged
                    ]

    if local_only:
        return local_tagged
    if include_local:
        return [*shared_tagged, *local_tagged]
    return shared_tagged


def _load_issues_from_directory(issues_dir: Path) -> List[IssueData]:
    return [
        read_issue_from_file(path)
        for path in sorted(issues_dir.glob("*.json"), key=lambda item: item.name)
    ]


def _tag_issue_source(issue: IssueData, source: str) -> IssueData:
    custom = {**issue.custom, "source": source}
    return issue.model_copy(update={"custom": custom})


def _tag_issue_project(issue: IssueData, root: Path, project_dir: Path) -> IssueData:
    project_path = _render_project_path(root, project_dir)
    custom = {**issue.custom, "project_path": project_path}
    return issue.model_copy(update={"custom": custom})


def _render_project_path(root: Path, project_dir: Path) -> str:
    root_resolved = root.resolve()
    project_resolved = project_dir.resolve()
    try:
        project_path = project_resolved.relative_to(root_resolved)
    except ValueError:
        project_path = project_resolved
    return str(project_path)


def _blocked_by_dependency(issue: IssueData) -> bool:
    return any(
        dependency.dependency_type == "blocked-by" for dependency in issue.dependencies
    )


def _validate_dependency_type(dependency_type: str) -> None:
    if dependency_type not in ALLOWED_DEPENDENCY_TYPES:
        raise DependencyError("invalid dependency type")


def _has_dependency(issue: IssueData, target_id: str, dependency_type: str) -> bool:
    return any(
        dependency.target == target_id and dependency.dependency_type == dependency_type
        for dependency in issue.dependencies
    )


def _ensure_no_cycle(project_dir: Path, source_id: str, target_id: str) -> None:
    graph = _build_dependency_graph(project_dir)
    graph.edges.setdefault(source_id, []).append(target_id)
    if _detect_cycle(graph, source_id):
        raise DependencyError("cycle detected")


def _build_dependency_graph(project_dir: Path) -> DependencyGraph:
    issues_dir = project_dir / "issues"
    edges: dict[str, list[str]] = {}
    for issue_path in issues_dir.glob("*.json"):
        issue = read_issue_from_file(issue_path)
        blocked_targets = [
            dependency.target
            for dependency in issue.dependencies
            if dependency.dependency_type == "blocked-by"
        ]
        if blocked_targets:
            edges[issue.identifier] = blocked_targets
    return DependencyGraph(edges=edges)


def _detect_cycle(graph: DependencyGraph, start: str) -> bool:
    visited: set[str] = set()
    stack: set[str] = set()

    def visit(node: str) -> bool:
        if node in stack:
            return True
        if node in visited:
            return False
        visited.add(node)
        stack.add(node)
        for neighbor in graph.edges.get(node, []):
            if visit(neighbor):
                return True
        stack.remove(node)
        return False

    return visit(start)



from __future__ import annotations

from pathlib import Path
from types import SimpleNamespace

from click.testing import CliRunner
import pytest

from kanbus import cli
from kanbus.content_validation import ContentValidationError
from kanbus.issue_close import IssueCloseError
from kanbus.issue_creation import IssueCreationError
from kanbus.issue_lookup import IssueLookupError
from kanbus.issue_transfer import IssueTransferError
from kanbus.issue_update import IssueUpdateError
from kanbus.migration import MigrationError

from test_helpers import build_issue, build_project_configuration


def _run(args: list[str]) -> object:
    return CliRunner().invoke(cli.cli, args)


def test_create_command_paths(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    monkeypatch.setattr(cli.Path, "cwd", lambda: tmp_path)
    monkeypatch.setattr(cli, "_run_lifecycle_hooks_for_context", lambda *_a, **_k: None)
    monkeypatch.setattr(cli, "format_issue_for_display", lambda *_a, **_k: "formatted")

    assert _run(["create"]).exit_code != 0
    assert _run(["create", "x", "--focus"]).exit_code != 0

    monkeypatch.setattr(
        cli,
        "apply_text_quality_signals",
        lambda text: SimpleNamespace(text=text, warnings=[], suggestions=[]),
    )
    monkeypatch.setattr(
        cli,
        "validate_code_blocks",
        lambda _text: (_ for _ in ()).throw(ContentValidationError("bad block")),
    )
    result_validate_fail = _run(["create", "x", "--description", "```json\n{\n```"])
    assert result_validate_fail.exit_code != 0
    assert "bad block" in result_validate_fail.output

    monkeypatch.setattr(cli, "validate_code_blocks", lambda _text: None)

    result_beads_local = _run(["--beads", "create", "x", "--local"])
    assert result_beads_local.exit_code != 0
    assert "does not support local issues" in result_beads_local.output

    issue = build_issue("kanbus-1")
    monkeypatch.setattr(cli, "create_beads_issue", lambda **_k: issue)
    monkeypatch.setattr(cli, "emit_signals", lambda *_a, **_k: None)
    result_beads_ok = _run(["--beads", "create", "x"])
    assert result_beads_ok.exit_code == 0
    assert "formatted" in result_beads_ok.output

    monkeypatch.setattr(
        cli,
        "create_beads_issue",
        lambda **_k: (_ for _ in ()).throw(cli.BeadsWriteError("beads create fail")),
    )
    result_beads_fail = _run(["--beads", "create", "x"])
    assert result_beads_fail.exit_code != 0
    assert "beads create fail" in result_beads_fail.output

    monkeypatch.setattr(
        cli,
        "create_issue",
        lambda **_k: SimpleNamespace(
            issue=issue, configuration=build_project_configuration()
        ),
    )
    result_regular_ok = _run(["create", "x", "--no-validate"])
    assert result_regular_ok.exit_code == 0

    monkeypatch.setattr(
        cli,
        "create_issue",
        lambda **_k: (_ for _ in ()).throw(IssueCreationError("create fail")),
    )
    result_regular_fail = _run(["create", "x", "--no-validate"])
    assert result_regular_fail.exit_code != 0
    assert "create fail" in result_regular_fail.output


def test_show_command_paths(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    monkeypatch.setattr(cli.Path, "cwd", lambda: tmp_path)
    monkeypatch.setattr(cli, "_run_lifecycle_hooks_for_context", lambda *_a, **_k: None)
    monkeypatch.setattr(cli, "format_issue_for_display", lambda *_a, **_k: "shown")

    issue = build_issue("kanbus-1")

    monkeypatch.setattr(
        cli,
        "load_project_configuration",
        lambda _p: build_project_configuration(beads_compatibility=True),
    )
    monkeypatch.setattr(
        cli, "get_configuration_path", lambda _p: tmp_path / ".kanbus.yml"
    )
    monkeypatch.setattr(cli, "_resolve_beads_root", lambda _r: tmp_path)
    monkeypatch.setattr(cli, "load_beads_issue", lambda _r, _i: issue)

    result_json = _run(["show", "kanbus-1", "--json"])
    assert result_json.exit_code == 0
    assert '"id": "kanbus-1"' in result_json.output

    result_text = _run(["show", "kanbus-1"])
    assert result_text.exit_code == 0
    assert "shown" in result_text.output

    monkeypatch.setattr(
        cli,
        "load_beads_issue",
        lambda *_a, **_k: (_ for _ in ()).throw(MigrationError("not found")),
    )
    result_beads_fail = _run(["show", "kanbus-1"])
    assert result_beads_fail.exit_code != 0

    monkeypatch.setattr(
        cli,
        "load_project_configuration",
        lambda _p: build_project_configuration(beads_compatibility=False),
    )
    monkeypatch.setattr(
        cli, "load_issue_from_project", lambda _r, _i: SimpleNamespace(issue=issue)
    )
    monkeypatch.setattr(
        cli, "get_configuration_path", lambda _p: tmp_path / ".kanbus.yml"
    )
    monkeypatch.setattr(
        "kanbus.console_snapshot.get_issues_for_root",
        lambda _r: (_ for _ in ()).throw(RuntimeError("ignore")),
    )
    result_regular_ok = _run(["show", "kanbus-1"])
    assert result_regular_ok.exit_code == 0

    monkeypatch.setattr(
        cli,
        "load_issue_from_project",
        lambda *_a, **_k: (_ for _ in ()).throw(IssueLookupError("lookup fail")),
    )
    result_regular_fail = _run(["show", "kanbus-1"])
    assert result_regular_fail.exit_code != 0


def test_update_command_paths(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    monkeypatch.setattr(cli.Path, "cwd", lambda: tmp_path)
    monkeypatch.setattr(cli, "_run_lifecycle_hooks_for_context", lambda *_a, **_k: None)
    monkeypatch.setattr(
        cli, "format_issue_key", lambda identifier, project_context=False: identifier
    )
    monkeypatch.setattr(cli, "emit_signals", lambda *_a, **_k: None)
    monkeypatch.setattr(
        cli,
        "apply_text_quality_signals",
        lambda text: SimpleNamespace(text=text, warnings=[], suggestions=[]),
    )

    monkeypatch.setattr(
        cli,
        "validate_code_blocks",
        lambda _text: (_ for _ in ()).throw(ContentValidationError("bad update")),
    )
    result_validate_fail = _run(["update", "kanbus-1", "--description", "x"])
    assert result_validate_fail.exit_code != 0

    monkeypatch.setattr(cli, "validate_code_blocks", lambda _text: None)

    result_beads_parent_fail = _run(["--beads", "update", "kanbus-1", "--parent", "p1"])
    assert result_beads_parent_fail.exit_code != 0

    issue = build_issue("kanbus-1")
    monkeypatch.setattr(cli, "load_beads_issue", lambda _r, _i: issue)
    monkeypatch.setattr(cli, "update_beads_issue", lambda *_a, **_k: None)
    result_beads_ok = _run(
        ["--beads", "update", "kanbus-1", "--set-labels", "a,b", "--no-validate"]
    )
    assert result_beads_ok.exit_code == 0
    assert "Updated kanbus-1" in result_beads_ok.output

    monkeypatch.setattr(
        cli,
        "update_beads_issue",
        lambda *_a, **_k: (_ for _ in ()).throw(
            cli.BeadsWriteError("beads update fail")
        ),
    )
    result_beads_fail = _run(["--beads", "update", "kanbus-1", "--no-validate"])
    assert result_beads_fail.exit_code != 0

    monkeypatch.setattr(
        cli,
        "load_project_configuration",
        lambda _p: (_ for _ in ()).throw(cli.ProjectMarkerError("pm")),
    )
    monkeypatch.setattr(
        cli, "load_issue_from_project", lambda _r, _i: SimpleNamespace(issue=issue)
    )
    monkeypatch.setattr(cli, "update_issue", lambda **_k: issue)
    result_regular_ok = _run(["update", "kanbus-1", "--claim", "--no-validate"])
    assert result_regular_ok.exit_code == 0

    monkeypatch.setattr(
        cli,
        "update_issue",
        lambda **_k: (_ for _ in ()).throw(IssueUpdateError("update fail")),
    )
    result_regular_fail = _run(["update", "kanbus-1", "--no-validate"])
    assert result_regular_fail.exit_code != 0


def test_close_move_promote_localize_comment_paths(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    monkeypatch.setattr(cli.Path, "cwd", lambda: tmp_path)
    monkeypatch.setattr(cli, "_run_lifecycle_hooks_for_context", lambda *_a, **_k: None)
    monkeypatch.setattr(
        cli, "format_issue_key", lambda identifier, project_context=False: identifier
    )
    issue = build_issue("kanbus-1")

    monkeypatch.setattr(cli, "close_issue", lambda _r, _i: issue)
    result_close = _run(["close", "kanbus-1"])
    assert result_close.exit_code == 0
    assert "Closed kanbus-1" in result_close.output

    monkeypatch.setattr(cli, "_resolve_beads_root", lambda _r: tmp_path)
    close_lookup_calls = {"count": 0}

    def _load_beads_issue_for_close(*_a, **_k):
        close_lookup_calls["count"] += 1
        if close_lookup_calls["count"] == 1:
            raise MigrationError("missing before")
        return issue

    monkeypatch.setattr(cli, "load_beads_issue", _load_beads_issue_for_close)
    monkeypatch.setattr(cli, "update_beads_issue", lambda *_a, **_k: None)
    result_close_beads = _run(["--beads", "close", "kanbus-1"])
    assert result_close_beads.exit_code == 0

    monkeypatch.setattr(
        cli,
        "update_beads_issue",
        lambda *_a, **_k: (_ for _ in ()).throw(MigrationError("close beads fail")),
    )
    result_close_beads_fail = _run(["--beads", "close", "kanbus-1"])
    assert result_close_beads_fail.exit_code != 0
    assert "close beads fail" in result_close_beads_fail.output

    monkeypatch.setattr(
        cli,
        "close_issue",
        lambda *_a, **_k: (_ for _ in ()).throw(IssueCloseError("close fail")),
    )
    assert _run(["close", "kanbus-1"]).exit_code != 0

    monkeypatch.setattr(
        cli,
        "load_project_configuration",
        lambda _p: build_project_configuration(beads_compatibility=False),
    )
    monkeypatch.setattr(
        cli, "get_configuration_path", lambda _p: tmp_path / ".kanbus.yml"
    )
    monkeypatch.setattr(cli, "update_issue", lambda **_k: issue)
    result_move = _run(["move", "kanbus-1", "bug"])
    assert result_move.exit_code == 0

    monkeypatch.setattr(
        cli,
        "load_project_configuration",
        lambda _p: build_project_configuration(beads_compatibility=True),
    )
    assert _run(["move", "kanbus-1", "bug"]).exit_code != 0

    monkeypatch.setattr(
        cli,
        "update_issue",
        lambda **_k: (_ for _ in ()).throw(IssueUpdateError("move fail")),
    )
    monkeypatch.setattr(
        cli,
        "load_project_configuration",
        lambda _p: build_project_configuration(beads_compatibility=False),
    )
    assert _run(["move", "kanbus-1", "bug"]).exit_code != 0

    monkeypatch.setattr(cli, "promote_issue", lambda *_a: None)
    monkeypatch.setattr(cli, "localize_issue", lambda *_a: None)
    monkeypatch.setattr(
        cli, "load_issue_from_project", lambda _r, _i: SimpleNamespace(issue=issue)
    )
    assert _run(["promote", "kanbus-1"]).exit_code == 0
    assert _run(["localize", "kanbus-1"]).exit_code == 0

    calls = {"count": 0}

    def _lookup_once_then_fail(_r, _i):
        calls["count"] += 1
        if calls["count"] == 1:
            return SimpleNamespace(issue=issue)
        raise IssueLookupError("missing after")

    monkeypatch.setattr(cli, "load_issue_from_project", _lookup_once_then_fail)
    assert _run(["promote", "kanbus-1"]).exit_code == 0

    calls["count"] = 0
    monkeypatch.setattr(cli, "load_issue_from_project", _lookup_once_then_fail)
    assert _run(["localize", "kanbus-1"]).exit_code == 0

    monkeypatch.setattr(
        cli,
        "promote_issue",
        lambda *_a: (_ for _ in ()).throw(IssueTransferError("promote fail")),
    )
    assert _run(["promote", "kanbus-1"]).exit_code != 0

    monkeypatch.setattr(
        cli,
        "localize_issue",
        lambda *_a: (_ for _ in ()).throw(IssueTransferError("localize fail")),
    )
    assert _run(["localize", "kanbus-1"]).exit_code != 0

    monkeypatch.setattr(
        cli,
        "apply_text_quality_signals",
        lambda text: SimpleNamespace(text=text, warnings=[], suggestions=[]),
    )
    monkeypatch.setattr(cli, "validate_code_blocks", lambda _t: None)
    monkeypatch.setattr(cli, "emit_signals", lambda *_a, **_k: None)
    monkeypatch.setattr(
        cli,
        "load_project_configuration",
        lambda _p: build_project_configuration(beads_compatibility=True),
    )
    monkeypatch.setattr(
        cli, "get_configuration_path", lambda _p: tmp_path / ".kanbus.yml"
    )
    monkeypatch.setattr("kanbus.beads_write.add_beads_comment", lambda *_a, **_k: None)
    monkeypatch.setattr(cli, "load_beads_issue", lambda *_a, **_k: issue)
    body_file = tmp_path / "comment.txt"
    body_file.write_text("from file", encoding="utf-8")
    result_comment_body_file = _run(
        ["comment", "kanbus-1", "--body-file", str(body_file), "--no-validate"]
    )
    assert result_comment_body_file.exit_code == 0

    monkeypatch.setattr(
        cli,
        "load_project_configuration",
        lambda _p: (_ for _ in ()).throw(cli.ProjectMarkerError("pm")),
    )
    monkeypatch.setattr(
        cli,
        "add_comment",
        lambda **_k: SimpleNamespace(issue=issue, comment=SimpleNamespace(id="c1")),
    )
    result_comment = _run(["comment", "kanbus-1", "hello"])
    assert result_comment.exit_code == 0

    monkeypatch.setattr(
        cli,
        "validate_code_blocks",
        lambda _t: (_ for _ in ()).throw(ContentValidationError("bad comment")),
    )
    assert _run(["comment", "kanbus-1", "hello"]).exit_code != 0


def test_update_beads_policy_signal_and_delete_beads_compat_paths(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    monkeypatch.setattr(cli.Path, "cwd", lambda: tmp_path)
    monkeypatch.setattr(cli, "_run_lifecycle_hooks_for_context", lambda *_a, **_k: None)
    monkeypatch.setattr(
        cli,
        "load_project_configuration",
        lambda _p: build_project_configuration(beads_compatibility=True),
    )
    monkeypatch.setattr(
        cli, "get_configuration_path", lambda _p: tmp_path / ".kanbus.yml"
    )
    monkeypatch.setattr(cli, "_resolve_beads_root", lambda _r: tmp_path)
    monkeypatch.setattr(
        cli,
        "apply_text_quality_signals",
        lambda text: SimpleNamespace(text=text.strip(), warnings=[], suggestions=[]),
    )
    monkeypatch.setattr(cli, "validate_code_blocks", lambda _t: None)
    monkeypatch.setattr(
        "kanbus.project.load_project_directory",
        lambda _root: tmp_path / "project",
    )
    (tmp_path / "project" / "policies").mkdir(parents=True)
    monkeypatch.setattr(
        "kanbus.policy_loader.load_policies", lambda _p: [("p.policy", object())]
    )
    monkeypatch.setattr(
        "kanbus.policy_evaluator.evaluate_policies", lambda *_a, **_k: []
    )
    monkeypatch.setattr(
        "kanbus.workflows.validate_status_value", lambda *_a, **_k: None
    )
    monkeypatch.setattr(
        "kanbus.workflows.validate_status_transition", lambda *_a, **_k: None
    )

    before_issue = build_issue("kanbus-1", labels=["keep", "drop"])
    load_calls = {"count": 0}

    def _load_beads_issue(*_a, **_k):
        load_calls["count"] += 1
        if load_calls["count"] == 1:
            return before_issue
        return build_issue(
            "kanbus-1",
            status="done",
            priority=1,
            labels=["a", "c"],
            custom=before_issue.custom,
        )

    monkeypatch.setattr(cli, "load_beads_issue", _load_beads_issue)
    monkeypatch.setattr(cli, "load_beads_issues", lambda _r: [before_issue])
    monkeypatch.setattr(cli, "update_beads_issue", lambda *_a, **_k: None)
    monkeypatch.setattr(
        cli, "format_issue_key", lambda identifier, project_context=False: identifier
    )
    emitted: list[str] = []
    monkeypatch.setattr(cli, "emit_signals", lambda *_a, **_k: emitted.append("emit"))

    result_update = _run(
        [
            "update",
            "kanbus-1",
            "--title",
            "Updated",
            "--description",
            "  desc  ",
            "--status",
            "done",
            "--priority",
            "1",
            "--assignee",
            "dev",
            "--set-labels",
            "a,b",
            "--add-label",
            "c",
            "--remove-label",
            "b",
        ]
    )
    assert result_update.exit_code == 0
    assert "Updated kanbus-1" in result_update.output
    assert emitted

    monkeypatch.setattr(
        cli,
        "format_issue_key",
        lambda identifier, project_context=False: identifier,
    )
    monkeypatch.setattr(
        cli,
        "load_beads_issue",
        lambda *_a, **_k: (_ for _ in ()).throw(MigrationError("missing")),
    )
    monkeypatch.setattr(cli, "delete_beads_issue", lambda *_a, **_k: None)
    result_delete = _run(["delete", "kanbus-1", "--yes"])
    assert result_delete.exit_code == 0
    assert "Deleted kanbus-1" in result_delete.output

    monkeypatch.setattr(
        "kanbus.beads_write.add_beads_comment",
        lambda *_a, **_k: None,
    )
    monkeypatch.setattr(cli, "load_beads_issue", lambda *_a, **_k: before_issue)
    result_comment = _run(["comment", "kanbus-1", "hello", "--no-validate"])
    assert result_comment.exit_code == 0
